/*
 * the core of kmouse -- offer vc-independent pointer support as devices 
 *
 * Copyright (C) 1996 Alessandro Rubini (rubini@linux.it)
 *
 * Several fixes by Markus Gutschke (gutschk@math.uni-muenster.de)
 * Good suggestions, termios ioctl faking, and Metro-X support by Aaron Ucko
 * (ucko@vax1.rockhurst.edu)
 *
 * Part of protocol decoding comes from selection, by Andrew Haylett
 * The idea of the kernel module is by Peter Orbaek
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 ********/

#ifndef __KERNEL__
#define __KERNEL__
#endif

/* By now, it's only a module. Time will tell... */

#ifndef MODULE
#define MODULE
#endif

#include <linux/module.h>

#if 0 /* the hard way to avoid "rmmod: device busy" :-) */
#  undef  MOD_INC_USE_COUNT
#  define MOD_INC_USE_COUNT
#  undef  MOD_DEC_USE_COUNT
#  define MOD_DEC_USE_COUNT
#endif 

#include <linux/types.h>     /* ulong and friends */
#include <linux/sched.h>     /* current, task_struct, other goodies */
#include <linux/fcntl.h>     /* O_NONBLOCK etc. */
#include <linux/errno.h>     /* return values */
#include <linux/config.h>    /* system name and global items */
#include <linux/malloc.h>    /* kmalloc, kfree */
#include <linux/kdev_t.h>    /* building majors */
#include <linux/major.h>     /* checking who's speaking */
#include <linux/tqueue.h>    /* running task queues */
#include <linux/tty.h>       /* line disciplines */
#include <linux/termios.h>   /* N_MOUSE -- it was already there... */
#include <linux/proc_fs.h>   /* debugging issues */
#include <linux/interrupt.h> /* mark_bh() */

#define __KERNEL_SYSCALLS__
#include <linux/unistd.h>  /* kernel_thread() */
#include <asm/unistd.h>

#include <asm/segment.h>

#include <linux/miscdevice.h>
#include <linux/../../drivers/char/selection.h> /* Hmmm.... not nice */

#include <gpm.h> /* gpm is needed, at least to define the interface */

#include "kmouse.h" /* kmouse interface */
#include "kmouse-config.h"

/* this is defined in misc.c */
extern int console_get_info(int *mode, int *shift, int *col, int *row,
			    struct tty_struct **tty);

static int minor=0; /* this is meant to be modified at load time */

#ifndef VERSION_CODE
#  define VERSION_CODE(vers,rel,seq) ( ((vers)<<16) | ((rel)<<8) | (seq) )
#endif

#ifdef KMOUSE_DEBUG
#define inline
#endif

#ifndef min
#define min(a,b) ((a)<(b)?(a):(b))
#endif

#if 0
#define kmalloc(a,b) ({ \
                      PDEBUG("%s:%i %i %i\n",__FILE__,__LINE__,a,b); \
                      kmalloc(a,b);})
@define kfree(a)     ({ \
                      PDEBUG("%s:%i free %i\n",__FILE__,__LINE__,a); \
                      kfree(a);})
#endif
	  
/*============================================== silly integer values */

static int statusC=0; /* clicks */
static int con_mode, con_fg, con_shift, con_wid, con_hei; /* console info */
static struct tty_struct *con_tty;

/* These are used to scale absolute pointers to relative */

static int range_x = 1024; /* these values can be chaged with ioctl() */
static int range_y = 768;

/*===================================================== inproto lists */

/*
 * in-protocols can be registered, unregistered and looked up
 */

struct kmouse_inproto *kmouse_inprotolist = NULL;

int kmouse_register_inproto(struct kmouse_inproto *ptr)
{
    ptr->next = kmouse_inprotolist;
    kmouse_inprotolist = ptr;
    PDEBUG("registering inproto %s\n",ptr->name);
    return 0;
}

int kmouse_unregister_inproto(struct kmouse_inproto *ptr)
{
    struct kmouse_inproto **ptrptr = &kmouse_inprotolist;

    if (ptr->usage) return -EBUSY;

    if (!ptr) return 0;
    while (*ptrptr && *ptrptr != ptr)
	ptrptr = &((*ptrptr)->next);

    if (!*ptrptr) {
	printk(KERN_WARNING "unregistering unexistent proto %s\n",
	       ptr->name);
	return -ENODEV;
    }

    *ptrptr = ptr->next;
    return 0;
}

struct kmouse_inproto *kmouse_lookup_inproto (char *name)
{
    struct kmouse_inproto *ptr = kmouse_inprotolist;
    while (ptr) {
        if (!strcmp(ptr->name, name))
            break;
        ptr = ptr->next;
    }
    return ptr;
}

/*===================================================== mouse options */
/* options are all relative to the current mouse */

#define opt_sequence (mPtr->opt->sequence)

#define opt_three (mPtr->opt->three)
#define DEFAULT_THREE 0

#define opt_delta (mPtr->opt->delta)
#define DEFAULT_DELTA 25

#define opt_accel (mPtr->opt->accel)
#define DEFAULT_ACCEL 2

#define opt_scale (mPtr->opt->scale)
#define opt_scaley (mPtr->opt->scaley)
#define DEFAULT_SCALE 10

#define opt_time (mPtr->opt->time)
#define DEFAULT_TIME 250

#define opt_tap (mPtr->opt->tap) /* unused, yet */
#define DEFAULT_TAP 1

#define opt_ptrdrag (mPtr->opt->ptrdrag)
#define DEFAULT_PTRDRAG 1

#define opt_chordmiddle (mPtr->opt->chordmiddle)
#define DEFAULT_CHORDMIDDLE 0

/*===================================================== mouse lists */

struct kmouse_mouse *kmouse_mlist = NULL;     /* list of registered devices */
struct kmouse_mouse *kmouse_lastmouse = NULL; /* cache the last one */

static struct kmouse_options default_opts = {
    0,                /* dev_t */
    "01234567",       /* sequence */
    0,                /* proto */
    DEFAULT_THREE,
    DEFAULT_DELTA,
    DEFAULT_ACCEL,
    DEFAULT_SCALE,
    DEFAULT_SCALE,
    DEFAULT_TIME,
    DEFAULT_TAP,
    DEFAULT_PTRDRAG,
    DEFAULT_CHORDMIDDLE
};

/*
 * When a new mouse spurs out, a data structure is allocated, and it is
 * initialized. What needs to be done is:
 *    giving default options
 *    initalizing the protocol, if known
 * Note that this initialization can be called more than once (if the
 * protocol is changed, for example)
 */

void kmouse_mouse_init(struct kmouse_mouse *mPtr, char *proto) {
    struct kmouse_inproto *protoptr;

    PDEBUGG("enter: mPtr %p, %p opt %p\n",mPtr,mPtr->proto,mPtr->opt);
    if (!mPtr->opt) {
        mPtr->opt = kmalloc(sizeof(struct kmouse_options),GFP_ATOMIC);
        if (mPtr->opt)
            *(mPtr->opt) = default_opts;
    }
    if (mPtr->opt) {
        opt_three = DEFAULT_THREE; /* restore this anyways */
        /*
         * the following operation is based on the observation that 80x50
         * has square cells. (An author-centric observation ;-)
         */
	if (con_wid)
	    opt_scaley = opt_scale * 50 * con_wid / 80 / con_hei;
    }
    protoptr = kmouse_lookup_inproto(proto);
    PDEBUGG("proto %s found at %p\n",proto,protoptr);
    if (protoptr == NULL) {
        printk(KERN_WARNING "kmouse: unknown protocol %s for dev %s\n",
               proto, kdevname(mPtr->dev));
    }
    if (protoptr != mPtr->proto) {
        if (mPtr->proto) {
            if (mPtr->proto->cleanup) mPtr->proto->cleanup(mPtr);
            mPtr->proto->usage--;
        }
        mPtr->proto = protoptr;
        if (protoptr) {
            if (protoptr->init)
                mPtr->xtrasize = protoptr->init(mPtr);
            mPtr->proto->usage++;
        }
    }
    PDEBUGG("leave: mPtr %p, %p opt %p\n",mPtr,mPtr->proto,mPtr->opt);
}

/*===================================================== reader's lists */

#define KMOUSE_FLAG_RAW   0x0100 /* don't calculate acc and delta */
#define KMOUSE_PROTO_MSC  0x0101 /* retrieve msc packets (raw) */
#define KMOUSE_PROTO_NEW  0x0002 /* retrieve the new protocol (not raw) */
#define KMOUSE_PROTO_RNW  0x0102 /* the raw new protocl (e.g., for X) */
#define KMOUSE_PROTO_GPM  0x0004 /* retrieve the gpm protocol (not raw) */

struct kmouse_client {
    struct file *filp;
    struct kmouse_client *next;   
    struct wait_queue *queue;
    union {
        struct {s32 *zptr; s32 *zbuf;} new; /* data for new protocol */
        struct {s32 x; s32 y;} msc;         /* data for msc reader */
	struct {u8 x,y,b;} gpm;             /* data for the gpm protocol */
    } u;
    int zpending;               /* pending bytes (new proto) */
    Gpm_Connect conn;
    unsigned short nitems;
    unsigned short proto;
    Kmouse_Event *first;                       /* where to dequeue from */
    Kmouse_Event *last;                        /* where to enqueue */
    unsigned char buf[sizeof(Kmouse_Packet)];  /* array of pending bytes */
    unsigned char pending;                     /* number of pending bytes */
};

struct kmouse_client *client_first=NULL;
struct kmouse_client *client_defhn=NULL; /* the first default handler */
struct kmouse_client *client_last=NULL;  /* last used -- chronologically */

struct termios kmouse_termios; /* to fake TCGETS -- AU */

/*
 * The following structure is used to communicate to/from the thread
 */
#define KMOUSE_FIFO_LEN 8
struct fifo_struct {
    enum {COPY, PASTE} action;
    unsigned short arg[5]; /* x1, y1, x2, y2, mode */
} kmouse_fifo[KMOUSE_FIFO_LEN];

int fifo_head, fifo_tail;

int kmouse_isqueued = 0;

/*============================================== managing the mouse list */

struct kmouse_mouse *kmouse_new=NULL;
struct wait_queue *kmouse_wait;

/*
 * Utility functions
 */

inline void kmouse_link(struct kmouse_mouse *mPtr)
{
    mPtr->head = mPtr->tail = 0;
    mPtr->missing = 2;
    mPtr->next = kmouse_mlist;
    kmouse_mlist = mPtr;
}

inline void kmouse_unlink(struct kmouse_mouse *mPtr)
{
    struct kmouse_mouse **ptrptr = &kmouse_mlist;
    PDEBUG("unink %p from %p\n",mPtr,kmouse_mlist);
    if (!mPtr) return;
    while (*ptrptr && *ptrptr != mPtr)
        ptrptr = &(*ptrptr)->next;
    if (!*ptrptr)
        return; /* not found */

    PDEBUGG("prev is %p, next is %p\n",ptrptr,mPtr->next);
    *ptrptr = mPtr->next;
    kfree(mPtr->opt);
    kfree(mPtr);
}

inline struct kmouse_mouse *kmouse_lookfor(kdev_t dev, int force)
{
    struct kmouse_mouse *mPtr=kmouse_mlist;

    if (kmouse_lastmouse && kmouse_lastmouse->dev == dev)
        return kmouse_lastmouse;

    while (mPtr && mPtr->dev != dev)
        mPtr = mPtr->next;

    if (!mPtr && force) {
        mPtr = kmouse_new;
        kmouse_new = NULL;
        wake_up_interruptible(&kmouse_wait); /* realloc it */
        if (!mPtr) return NULL; /* shouldn't happen */
        mPtr->dev = dev;
        kmouse_link(mPtr);
    }
    return kmouse_lastmouse = mPtr;
}

/*============================================== the thread stuff */

int kmouse_killed = 0; /* a global flag to tell to shut down */

/*
 * This is a kernel thread, in order to allow sleeping. This is
 * definitely needed to be able to select and to paste. While
 * we're at it, use the thread to allocate data as well
 */
int kmouse_thread(void *unused)
{
    struct kmouse_client *ptr;
    struct task_struct *p;
    struct kmouse_mouse *mPtr;
    void kmouse_close_devices(void);

    /* prepare the thread so it looks good (make it real-time by now) */
    current->session = 1;
    current->pgrp = 1;
    current->policy = SCHED_RR; /* real-time, round-robin */
    current->rt_priority = 0;
    sprintf(current->comm,"kmouse");

    /* what we need now is disassociate from userland */

    /*
     * If the mm structure is NULL, do_exit will oops().  The only mm
     * which is not freed is init_mm, which is static. the symbol is
     * not public, so assign current->mm to the one of process 2
     * (kflushd).
     */
    for (p = current ; (p = p->next_task) != current ; )
        if (p->pid == 2)
            break;
    current->mm->count--; /* detatch */
    current->mm = p->mm;

    /* don't attach, as init_mm doesn't use the count */

    current->fs->count--; /* incremented by CLONE_FS */
    current->fs=NULL;
    current->files->count--; /* incremented by CLONE_FILES */
    current->files=NULL;
#if 0
    /* can't zero current->sig, otherwise I can't be killed */
    current->sig->count; /* incremented by CLONE_SIGHAND */
    current->sig=NULL;
#endif
    if (current->exec_domain && current->exec_domain->use_count)
      (*current->exec_domain->use_count)--;
    if (current->binfmt && current->binfmt->use_count)
      (*current->binfmt->use_count)--;
    
    while (1) {
        interruptible_sleep_on(&kmouse_wait);
        if (current->signal & ~current->blocked) {
            PDEBUGG("incrementing killed\n");
            kmouse_killed++; /* tell everything to terminate */
            for (ptr = client_first; ptr; ptr = ptr->next)
                wake_up_interruptible(&ptr->queue); /* awake clients */
            break;
        }
	/*
	 * Here we are: now do the required task
	 */
	if (!kmouse_new) {
	    kmouse_new = kmalloc(sizeof(struct kmouse_mouse), GFP_KERNEL);
	    memset(kmouse_new,0,sizeof(struct kmouse_mouse));
	}

	while (fifo_head != fifo_tail) {
	    if (kmouse_fifo[fifo_tail].action == PASTE) {
                if (con_tty)
                    paste_selection((struct tty_struct *)con_tty);
            }
            else
		set_selection((long)(kmouse_fifo[fifo_tail].arg)-1,
			      NULL /* tty unused */, 0 /* kernel */);
	    fifo_tail++;
	    fifo_tail &= KMOUSE_FIFO_LEN -1;
	}
    }
    /*
     * Ok, if the thread was killed, then kmouse is unloading.
     * Close the devices (and thus the line discipline as well),
     * and disallocate the mice. Other cleanups are then
     * performed by cleanup_module(), once the usage is 0.
     */

    kmouse_close_devices();
    for (mPtr = kmouse_mlist; mPtr; mPtr = kmouse_mlist) {
        kmouse_unlink(mPtr);
    }

    PDEBUGG("thread killed\n");
    MOD_DEC_USE_COUNT; /* the thread counted as usage */
    return 0;
}
	   

/*============================================== decoding the protocol */
/*
 * eptr->zptr is NULL on entry. Protocol decoding allocates it.
 */

inline void kmouse_decode(struct kmouse_mouse *mPtr,
                                 Kmouse_Event *eptr)
{
    int n=mPtr->head-mPtr->tail;        /* bytes to process */
    static unsigned long tv1, tv2;      /* jiffies, to check double clicks */
    static short newB=0, oldB=0;        /* old values to chain events */
    static unsigned long ljiffies;      /* last invocation */
    struct kmouse_inproto *p;
    int eaten, i;

    /*
     * First of all, try to autodetect the protocol, if none has been
     * specified by the user
     */
#ifdef KMOUSE_DEBUG
    if (0) {
        PDEBUG("data (%i): %02x",n, mPtr->data[mPtr->tail] & 0xff);
        for (i=1; mPtr->tail+i < mPtr->head; i++)
            printk(" %02x",mPtr->data[mPtr->tail+i] & 0xff);
        printk("\n");
    }
#endif

    if (mPtr->errors > 10 && mPtr->errors > mPtr->decoded) {
        PDEBUG("too many errors. Trying autodetect\n");
        if (mPtr->proto) mPtr->proto->usage--;
        mPtr->proto = NULL;
        mPtr->decoded = mPtr->errors = 0;
    }
    if (mPtr->proto == NULL) { /* new mouse: autodetect */
        if (n<2) {
            mPtr->missing=2-n;
            return;
        }
        ljiffies=jiffies;
        for (p = kmouse_inprotolist; p; p = p->next) {
            if ( (mPtr->data[mPtr->tail]&p->proto[0])
                != p->proto[1])
                continue;
            if ( (mPtr->data[mPtr->tail+1]&p->proto[2])
                != p->proto[3])
                continue;
            PDEBUG("found proto: %s (%i)\n",p->name, p->protoid);
	    /*
	     * Ok, now init the mouse with the new proto. Initialization
	     * uses the name of the proto, so ioctl() can pass it 
	     */
	    kmouse_mouse_init(mPtr, p->name);
            mPtr->missing=p->howmany - n;
            break;
        }
	if (!p) { /* not found: skip one byte and loop */
	    mPtr->tail++;
	    return;
	}
    }
    
    /*
     * Ok, here we are, now decode all the packets in the list
     */
    
    if (n >= 2*mPtr->proto->howmany) PDEBUG("%i bytes to eat\n",n);

    /* while there is a whole packet, or the first byte is an extra */
    while (n) {
        /*
         * extra byte (damned mman protocol). Isn't it possible to
         * force it to msc?
         */
        if ( (mPtr->data[mPtr->tail] & mPtr->proto->extra[0])
             == mPtr->proto->extra[1]) {
            if (!mPtr->decoded) {
                mPtr->tail++; n--; /* first packet: skip leading bytes */
                continue;
            }
            eaten = (*mPtr->proto->fun)(eptr, 
                                        &(mPtr->data[mPtr->tail]),mPtr);
            if (eaten < 1) mPtr->tail++; n--; continue; /* not right */
            mPtr->tail++; n--; /* eaten: skip it and go on */
        }
        else if (n < mPtr->proto->howmany) break; /* too few bytes */
        else {
	    /*
	     * So, try to decode a whole packet.
	     * First, check the magic numbers.
	     */
            if ( mPtr->proto->howmany >= 2 &&
                (mPtr->data[mPtr->tail] & mPtr->proto->proto[0])
                != mPtr->proto->proto[1]) {
                if (mPtr->data[mPtr->tail] || mPtr->decoded) {
                    PDEBUG("not head %2x? (e=%i,d=%i)\n",
                           mPtr->data[mPtr->tail],mPtr->errors,mPtr->decoded);
                    mPtr->errors++;
                }
                n--; mPtr->tail++; /* skip unaligned bytes */
                continue;
            }
            if ((mPtr->data[mPtr->tail+1] & mPtr->proto->proto[2])
                != mPtr->proto->proto[3]) {
                PDEBUG("data packet?\n");
                mPtr->errors++;
		n--; mPtr->tail++; /* skip a single byte, and retry */
                continue;
            }

	    /*
             * the function returns how much it ate up. If there
             * are Z dimensions, they have been allocated by it.
             */
            eaten = (*mPtr->proto->fun)(eptr, &(mPtr->data[mPtr->tail]), mPtr);
	    if ( (i = eaten<1) ) {
                eaten = -eaten; /* no data, but skip bytes */
                if (!eaten) {
                    PDEBUG("Wrong protocol returned 0\n");
		    mPtr->errors++;
                    eaten++;
                }
            }
            mPtr->tail += eaten;
            n -= eaten;
	    if (i) {
                PDEBUG("dropping %i bytes\n",eaten);
                continue;
            }
            mPtr->decoded++;  /* eaten was >0 */
        }

        /* -------------------------------------------------
	 * hw decoding is over: the valid fields in event are now
	 * p.x, p.y, p.z (Hmm...) p.flags, p.buttons. 
         * Still missing: p.magic, p.shift, vc, clicks, type, margin, opt
         * and the dx/dy material for gpm events...
         */
    
        con_fg=console_get_info(&con_mode, &con_shift,
                                &con_wid, &con_hei, &con_tty);

        eptr->p.shift = con_shift;
        eptr->vc = con_fg+1;
        eptr->p.magic = KMOUSE_MAGIC;
	eptr->opt = mPtr->opt;

	/*
	 * Reorder the first three buttons, leave unchanged the other ones
	 */

        eptr->p.buttons = (eptr->p.buttons & ~7) |
	    (opt_sequence[eptr->p.buttons&7] & 7); 

	/*
	 * If the middle has been used, then they are at least three
	 */
        if ((eptr->p.buttons & GPM_B_MIDDLE) && !opt_three) {
            PDEBUGG("incrementing three\n");
            opt_three++;
        }
    
        /*
         * still missing: type, clicks, margin. Fix the type
         */
        oldB = newB; newB = eptr->p.buttons;
        /* FIXME -- place to put button-lock */

#if 0
        if (!kmouse_is_absolute(&(eptr->p)) 
            && !eptr->p.x && !eptr->p.y && (newB==oldB)) {
            eptr->type=0;
            continue; /* nothing new to report to userspace */
        }
#endif    
        if (oldB == newB)
            eptr->type = (eptr->p.buttons ? GPM_DRAG : GPM_MOVE);
        else
            eptr->type = (eptr->p.buttons > oldB ? GPM_DOWN : GPM_UP);
    
        switch(eptr->type)          /* now provide the cooked bits */
            {
            case GPM_DOWN:
                tv2=jiffies;
                if (tv1 && (tv2-tv1)*1000/HZ < opt_time)
                    statusC++, statusC%=3; /* 0, 1 or 2 */
                else statusC=0;
                eptr->type|=(GPM_SINGLE<<statusC);
                break;
                
            case GPM_UP: /* FIXME -- still to tell which button */
                tv1=jiffies;
                /* fall through */
                
            case GPM_DRAG:
                eptr->type |= GPM_MFLAG; /* Hmmm... check this one */
                eptr->type |= (GPM_SINGLE<<statusC);
                break;
                
            case GPM_MOVE:
                statusC=0;

            default:
                break;
            }
        eptr->clicks=statusC;
    
        /*
         * Still missing: the margin. Do it in realxy(), which is only
         * called for gpm packets (selection and /dev/gpmctl).
         */
    } /* while (n) */
    mPtr->missing=mPtr->proto->howmany-n;
}


/*============================================== the selection mechanism */
/*
 * The function handling selection is called from kmouse_dispatch,
 * which in turn runs on the immediate queue.  Unfortunately,
 * selection dynamically allocats the buffer (that was me :) using a
 * non-atomical kmalloc. Thus, set_selection() and paste_selection()
 * are dispatched to the kmouse thread.
 */

static inline void km_set_selection(x1,y1,x2,y2,m)
{
    unsigned short *ptr = kmouse_fifo[fifo_head].arg;

    kmouse_fifo[fifo_head].action = COPY;
    *(ptr++) = x1;
    *(ptr++) = y1;
    *(ptr++) = x2;
    *(ptr++) = y2;
    *(ptr++) = m;
    fifo_head++;
    fifo_head &= KMOUSE_FIFO_LEN-1;
    if (fifo_head == fifo_tail)
	PDEBUG("selection-buffer overflow");
    wake_up_interruptible(&kmouse_wait);
}

static inline void km_paste_selection(void)
{
    kmouse_fifo[fifo_head].action = PASTE;
    fifo_head++;
    fifo_head &= KMOUSE_FIFO_LEN-1;
    if (fifo_head == fifo_tail)
	PDEBUG("selection-buffer overflow");
    wake_up_interruptible(&kmouse_wait);
}

/*============================================== char-cell coordinates */
static int screenX=1, screenY=1; /* the char-cell position */

void kmouse_realxy(Kmouse_Event *eptr)
{
    static int fine_dx=0, fine_dy=0; 
    /* allow to exit only if drag or up */
    int out = !!(eptr->type & (GPM_DRAG|GPM_UP));

    if (kmouse_absolute_p(&(eptr->p))) {
        /* 
         * Ok, here we are. When using char cells, it's important
         * to leave some place for out-of-screen area: this is used
         * to be able to easily select whole lines. Use 12% on each side
         * i.e.: multiply by 1.25 and subtract 0.125
         */
        screenX = (((eptr->p.x >> 16) * 5 - (1<<13)) * con_wid) >> 16;
        screenY = (((eptr->p.y >> 16) * 5 - (1<<13)) * con_hei) >> 16;
    } else {
        fine_dx += eptr->p.x;
        fine_dy += eptr->p.y;
        screenX += fine_dx / eptr->opt->scale;
        fine_dx %= eptr->opt->scale;
        screenY += fine_dy / eptr->opt->scaley;
        fine_dy %= eptr->opt->scaley;
    }
    eptr->margin = 0;
    if (screenX < 1) {
        screenX = 1 - out; if (out) eptr->margin |= GPM_LFT;
    }
    if (screenY < 1) {
        screenY = 1 - out; if (out) eptr->margin |= GPM_TOP;
    }
    if (screenX > con_wid) {
        screenX = con_wid + out; if (out) eptr->margin |= GPM_RGT;
    }
    if (screenY > con_hei) {
        screenY = con_hei + out; if (out) eptr->margin |= GPM_BOT;
    }
}

int do_selection(Kmouse_Event *eptr)  /* return 0, always */
{
    /*
     * The function gets a Kmouse_Event: it needs to recreate
     * the x and y fields. It uses also the buttons, type and
     * clicks fields. FIXME -- absolute pointers aren't handled yet
     */
    static int x1=1, y1=1; /* start at top-left, by now */
    static int x2, y2;
    /*
     * Both selecting and pasting is performed by the thread, 'cause they
     * might involve sleeping. 
     */
    kmouse_realxy(eptr);    
    PDEBUGG("abs %i: %i,%i -> %i,%i\n",!!kmouse_absolute_p(&(eptr->p)),
           eptr->p.x, eptr->p.y, screenX, screenY);

    x2=screenX; y2=screenY;
    
    switch(GPM_BARE_EVENTS(eptr->type))
    {
    case GPM_MOVE:
        if (x2<1) x2++; else if (x2>con_wid) x2--;
        if (y2<1) y2++; else if (y2>con_hei) y2--;
        km_set_selection(x2,y2,x2,y2,3); /* just highlight pointer */
        return 0;
        
    case GPM_DRAG:
        if (eptr->p.buttons==GPM_B_LEFT) {
            if (eptr->margin) /* fix margins */
                switch(eptr->margin) {
                  case GPM_TOP: x2=1; y2++; break;
                  case GPM_BOT: x2=con_wid; y2--; break;
                  case GPM_RGT: x2--; break;
                  case GPM_LFT: y2<=y1 ? x2++ : (x2=con_wid, y2--); break;
                }
            km_set_selection(x1,y1,x2,y2,eptr->clicks);

            if (eptr->clicks>=eptr->opt->ptrdrag && !eptr->margin)
                km_set_selection(x2,y2,x2,y2,3);  /* pointer */
        }
        return 0;
        
    case GPM_DOWN:
        switch (eptr->p.buttons) 
        {
        case GPM_B_LEFT:
            x1=x2; y1=y2;
            km_set_selection(x1,y1,x2,y2,eptr->clicks); /* start */
            return 0;
            
        case GPM_B_MIDDLE:
            /*
             * Hmm... pasting big chunks *can* cause sleeping. And we can
             * be in the idle task. Let's use the kmouse kernel thread.
             */
	    km_paste_selection();
            return 0;

        case GPM_B_RIGHT:
            PDEBUGG("three is %i\n",eptr->opt->three);
            if (eptr->opt->three)
                km_set_selection(x1,y1,x2,y2,eptr->clicks);
            else
                km_paste_selection();
            return 0;
        }
    }
    return 0;
}




/*============================================== enqueueing to a client */
/*
 * The function `eats' the packet, unless it is clustered.
 * The return value is NULL if the event was eaten, eptr otherwise
 */
inline Kmouse_Event *kmouse_enqueue(Kmouse_Event *eptr,
                                    struct kmouse_client *client)
{
    
    Kmouse_Event *last = client->last;
    int i, should_acc = 1;

    eptr->next = NULL; /* do it once for all */
    /*
     * First of all, apply  accel if (!raw && !abs)
     */
    if (!(client->proto & KMOUSE_FLAG_RAW)
        && !kmouse_absolute_p(&(eptr->p))
        && (abs(eptr->p.x)+abs(eptr->p.y) > eptr->opt->delta) )  {
        PDEBUG("accel\n");
        /* Don't do it now: if the event is not eaten, it is reused */
        should_acc = eptr->opt->accel;
    }
    
    if (!client->first) {
        client->first = client->last = eptr; 
        PDEBUGG("Feeding client %p\n",client);
        client->nitems++;
	eptr->p.x *= should_acc;
	eptr->p.y *= should_acc;
        PDEBUGG("zptr is %p\n",client->first.zptr);
        wake_up_interruptible(&client->queue);
        return NULL;
    }
    

    /* FIXME -- place to add cluster/nocluster control */

    /*
     * If clustering, don't eat the packet
     */
    if ( (last->type == eptr->type) /* buttons can't change within type */
        && (last->p.flags == eptr->p.flags)
        && (last->p.shift == eptr->p.shift)) {
        PDEBUGG("clustering (b %i->%i, t %x, f %x)\n",
                last->p.buttons,eptr->p.buttons,
                last->type, last->p.flags);

        if (kmouse_absolute_p(&(eptr->p))) {
            last->p.x = eptr->p.x;
            last->p.y = eptr->p.y;
            for (i = kmouse_nr_dimensions(&(eptr->p))-2; i>0; i--)
                last->zptr[i] = eptr->zptr[i];
        } else {
            last->p.x += eptr->p.x * should_acc;
	    last->p.x += eptr->p.x * should_acc;
            for (i = kmouse_nr_dimensions(&(eptr->p))-2; i>0; i--)
                last->zptr[i] += eptr->zptr[i];
        }
        return eptr; /* not eaten */
    } 
    /*
     * else, don't cluster: store it and return
     */
    last->next = eptr;
    client->nitems++;
    client->last = eptr;
    return NULL;
}

/*===================================================== dispatch to readers */
/*
 * This function runs in the immediate task queue, and dispatches collected
 * data to a client (a program reading from one of the devices). If no
 * clients are there, the selection mechanism is run.
 *
 * BUGS: shouldn't scan the whole list of clients every time: add some caching
 */
static struct tq_struct kmouse_tqueue;

void kmouse_dispatch (void *unused) /* this is called from the task_queue */
{
    struct kmouse_mouse *mPtr;
    struct kmouse_client *client;
    static Kmouse_Event event;
    static Kmouse_Event *saveevent; /* rebuilt any now and then */

    PDEBUGG("Running tqueue\n");

    /* scan the list and process data */

    if (!saveevent) 
        saveevent = kmalloc(sizeof(Kmouse_Event), GFP_ATOMIC);

    for (mPtr = kmouse_mlist; mPtr; mPtr = mPtr->next) {
        if (mPtr->head == mPtr->tail) /* empty */
            continue;
        if (event.zptr) {
            PDEBUGG("forgot a zptr, freeing it now\n");
            kfree(event.zptr);
            event.zptr = NULL;
        }
        kmouse_decode(mPtr,&event); /* on return zptr is allocated */
        if (mPtr->head == mPtr->tail) 
            mPtr->tail = mPtr->head = 0;
        PDEBUGG("left %i (m %i), (h,t) = (%i,%i)\n",n,mPtr->missing,
                mPtr->head,mPtr->tail);
        if (event.type == 0) continue; /* nothing */;
    
	/* scan the client list and awake them, or run selection */

	for (client=client_first; client; client=client->next) {
            if (client->conn.vc  && client->conn.vc != event.vc)
                continue; /* different console */
            if ((client->conn.minMod & event.p.shift) < client->conn.minMod)
                continue; /* too few modifiers */
            if ((client->conn.maxMod & event.p.shift) < event.p.shift)
                continue; /* too many of them */
        
            /* event type, considering the default mask as well */
            if ((client->conn.eventMask & event.type) == 0)
                if (client->conn.defaultMask & event.type)
                    continue; /* look for another client */
                else
                    break; /* drop this event */

            /*
             * Ok, we know the client.
             * Do it now, even if we're in the timer interrupt.
             */
            if (!saveevent)
                saveevent = kmalloc(sizeof(Kmouse_Event),GFP_ATOMIC);
            if (saveevent) {
                *saveevent = event;
                saveevent = kmouse_enqueue(saveevent,client);
            } else {
                PDEBUG("Can't allocate event for storing\n");
            }
            if ((client->conn.defaultMask & (event.type|GPM_HARD)) ==
                (event.type|GPM_HARD))
                continue;

            break; /* eaten */
        }

        if (!client)
            do_selection(&event);
        if (!saveevent)
            saveevent = kmalloc(sizeof(Kmouse_Event),GFP_ATOMIC);
    }
    /* done */
    kmouse_isqueued--; /* atomic? */
}

/*===================================================== stealing functions */

/*
 * These are called from within an interrupt during normal operation,
 * but from a write() call if faking mice (I've no busmouse to check).
 * The stolen bytes are what drive the whole mechanism
 */
void kmouse_steal_char(kdev_t dev, unsigned char datum)
{
    struct kmouse_mouse *mPtr;
    int i, j;

    /* save it somewhere */
    PDEBUGG("receiving %02x\n",datum);
    mPtr=kmouse_lookfor(dev, 1 /* alloc if needed */);
    if (!mPtr) return; /* shouldn't happen */
    if (!mPtr->proto && !datum && MAJOR(dev)!=MISC_MAJOR) {
        PDEBUG("ignoring leading zero\n"); /* Hmmm.... */
        return;
    }

    mPtr->data[mPtr->head++]=datum;
    mPtr->missing--;
    if (mPtr->head==KMOUSE_BUFFER_LEN) {
        if (mPtr->tail < 8) /* buffer tight */
            PDEBUG("Out of buffer, freeing %i\n",mPtr->tail);
        /*
         * Well, if the buffer is full, It's better to remove a whole
         * packet instead of a single byte (which would unalign the
         * data
         */
        if (!mPtr->tail)
            mPtr->tail = mPtr->proto->howmany;
        for (i=mPtr->tail, j=0; i<mPtr->head; i++, j++)
            mPtr->data[j]=mPtr->data[i];
        mPtr->head=j;
        mPtr->tail=0;
        mPtr->missing=0; kmouse_isqueued = 0; /* force registering */
    }        
    if (mPtr->missing<=0 && kmouse_isqueued<=0) {
        queue_task(&kmouse_tqueue, &tq_immediate);
	mark_bh(IMMEDIATE_BH);
        kmouse_isqueued=1;
    }
}

#if 0 /* FIXME -- steal-packet */

void kmouse_steal_packet( .............. )
{
    struct kmouse_inproto *p;

    PDEBUGG("stolen (%i,%03i,%03i) from %x\n",buttons,dx,dy,dev);

    /* enqueue three bytes, use sun protocol (minimum overhead) */
    kmouse_steal_char(dev,  (buttons & 0x07) | 0x80);

    if (kmouse_lastmouse->proto == NULL) { /* first time, tell it's sun */
        for (p=protocols; p->protoid != PROTO_SUN; p++)
            ;
        kmouse_lastmouse->proto = p;
    }
     
    if (dx>127) dx = 127; else if (dx < -127) dx = -127; /* dirty */
    if (dy>127) dy = 127; else if (dy < -127) dy = -127;
    kmouse_steal_char(dev, dx & 0xff);
    kmouse_steal_char(dev, dy & 0xff);
}

#else

void kmouse_steal_packet( kdev_t dev, int buttons, int flags,
			    int dx, int dy, int dz, int d4)
{}

#endif

/*===================================================== the line discipline */

struct tty_ldisc kmouse_ldisc;

int kmouse_ldisc_open(struct tty_struct *unused)
{
    MOD_INC_USE_COUNT;
    PDEBUGG("open ldisc, count is 0x%lx\n",mod_use_count_);
    return 0;
}

void kmouse_ldisc_close(struct tty_struct *unused)
{
    MOD_DEC_USE_COUNT;
    PDEBUGG("close ldisc, count is 0x%lx\n",mod_use_count_);
}

int kmouse_ldisc_room(struct tty_struct *unused)
{
    return 64; /* anything */
}

void kmouse_ldisc_receive(struct tty_struct *tty,
                                 const unsigned char *cp, char *fp, int count)
{
  int i=0;
  while (i<count)
      kmouse_steal_char(tty->device,cp[i++]);
}


/*==================================================== open&close ptr device */

struct kmouse_device {
    struct inode *inode;
    struct file *filp;
    struct kmouse_device *next;
};

struct kmouse_device *kmouse_devicelist = NULL;

int kmouse_open_device(char *name)
{
    struct kmouse_device *device;
    struct file *filp;
    struct inode *inode;
    int res;

    /*
     * Hmmm... can't do it with the current user pointer, as it
     * open the file for the current process.
     * So, open_namei readonly, alloc a new filp, call the fops.
     */


    if ( (res = open_namei(name, O_RDONLY, 0 /* mode */, &inode, NULL)) )
        return res;

    /* Check for duplicates: refuse to re-open the same device */
    res = -EBUSY;
    for (device = kmouse_devicelist; device; device = device->next)
        if (device->inode->i_rdev == inode->i_rdev)
            goto o_fail_icheck;

    res = -ENOMEM;
    if (!(device = kmalloc(sizeof(struct kmouse_device), GFP_KERNEL)))
        goto o_fail_icheck;
    if (!(filp = kmalloc(sizeof(struct file), GFP_KERNEL)))
        goto o_fail_filp;

    device->filp = filp;
    memset(filp, 0, sizeof(struct file));
    device->inode = inode;

    /* These lines from open.c */
    filp->f_inode = device->inode;
    filp->f_pos = 0;
    filp->f_reada = 0;
    filp->f_op = NULL;
    filp->f_flags = O_RDONLY;
    if (inode->i_op)
        filp->f_op = inode->i_op->default_file_ops;
    if (filp->f_op && filp->f_op->open)  res = filp->f_op->open(inode,filp);
    else res = 0; /* Hmmm... a device without an open method? */

    if (res == 0) {
        device->next = kmouse_devicelist;
        kmouse_devicelist = device;
	return 0;
    }

    kfree(filp);
  o_fail_filp: kfree(device);
  o_fail_icheck: iput(inode);
    return res;

}

void kmouse_close_devices(void)
{
    struct kmouse_device *device, *next;

    for (device = kmouse_devicelist; device; device = next) {
        PDEBUGG("release device %s\n", device->inode ?
	       kdevname(device->inode->i_rdev) : "<unknown>");
        device->filp->f_op->release(device->inode, device->filp);
        iput(device->inode);
        kfree(device->filp);
        next = device->next;
        kfree(device);
    }
    kmouse_devicelist = NULL;
}

/*===================================================== file operations */

/*
 * Silly lseek. Just deny it.
 */
int kmouse_seek(struct inode *inode, struct file *filp,
                       off_t offset, int origin)
{ return -ESPIPE;}

/*
 * The open method: identify the node, and treat X specially.
 */
int kmouse_open (struct inode *inode, struct file *filp)
{
    /* the f_op field is already right -- misc.c did it for us */
    struct kmouse_client *client;
    static int minortoproto(int minor);

    client=kmalloc(sizeof(struct kmouse_client),GFP_ATOMIC);
    if (!client) return -ENOMEM;
    memset(client, 0, sizeof(struct kmouse_client));
    client->proto = minortoproto(MINOR(inode->i_rdev));
    PDEBUGG("open: proto is %i\n",client->proto);
    /*
     * Ok, at this point the connection proto should be assigned:
     * the control node (/dev/gpmctl) gets nothing -- it must be cfg'd,
     * while the other get everything for the current console.
     */
    if (client->proto == KMOUSE_PROTO_GPM) {
        client->conn.vc = ~0;
    } else { /* a non-gpm node */
        int con = console_get_info(NULL,NULL,NULL,NULL,NULL);
        kdev_t device = MKDEV(TTY_MAJOR,con+1);
        /*
         * We need to treat X specially, because it switches ttys on us.
	 * This test used to be performed only when there was no
	 * controlling tty, which is true for XFree86 but not Metro-X.  I
	 * wish there were a better way to identify X servers; the present
	 * test is neither necessary nor sufficient :-( -- AU
	 */
        if (current->comm[0]=='X' && current->comm[1]=='\0') {
            PDEBUGG("X: uid %i, euid %i\n",current->uid, current->euid);
        } else {
            if (current->tty) {
                device = current->tty->device;
            } else if (!suser()) {
                PDEBUG("open: no ctrl tty\n");
                kfree(client);
                return -EPERM;
            } else {
                device = MKDEV(TTY_MAJOR,0); /* allow open, but no events */
            }
        }
        if (MAJOR(device) != TTY_MAJOR
            || MINOR(device) > MAX_NR_CONSOLES) {
            kfree(client);
            return -EPERM;
        }
        client->conn.vc = MINOR(device);
        PDEBUGG("tty is %i\n",MINOR(device));
        client->conn.maxMod = ~0;
        client->conn.eventMask = ~0;
        #if 0
        client->conn.minMod = 0;
        client->conn.defaultMask = 0;
        #endif
    }
    client->filp = filp;
    client->next = client_first;
    client_first = client;
    filp->private_data = client;


    MOD_INC_USE_COUNT;
    PDEBUGG("Opening at %p (%p) by %i (%s)\n",client,filp,
           current->pid,current->comm);
    PDEBUGG("count is 0x%lx\n",mod_use_count_);
    return 0;
}

/*
 * Closing the file
 */
void kmouse_release(struct inode *inode, struct file *filp)
{
    struct kmouse_client **ptrptr = &client_first;
    struct kmouse_client *client = filp->private_data;
    Kmouse_Event *ePtr, *nPtr;
    
    /* detach from the list */
    while (*ptrptr && *ptrptr != client)
        ptrptr=&(*ptrptr)->next;
    if (!*ptrptr) { /* pedantic */
        printk(KERN_ERR "kmouse: closing unopened node\n");
        return;
    }
    *ptrptr = client->next;

    PDEBUGG("Closing at %p (%p) by %i (%s)\n",client,filp,
           current->pid,current->comm);
    PDEBUGG("First event is %p\n",client->first);

    /* free it all */
    for (ePtr=client->first; ePtr; ePtr=nPtr) {
        PDEBUGG("first is %p, now is %p, next is %p\n",
               client->first, ePtr, ePtr->next);
        nPtr=ePtr->next;
        if (ePtr->zptr)
            kfree(ePtr->zptr);
        kfree(ePtr);
    }
    kfree(client);
    MOD_DEC_USE_COUNT;
    PDEBUGG("count is 0x%lx\n",mod_use_count_);
}

/*
 * ioctl. One of the commands opens the device from kernel mode. This
 * way the driving process can exit while leaving the settings in force.
 * In other words, gpm is no more needed to drive kmouse.
 * Unfortunately, this causes warning messages to be printed on the logs,
 * because our filp is not registered in the global list of files.
 */
int kmouse_ioctl(struct inode *inode, struct file * filp,
                         unsigned int cmd, ulong arg)
{
    int err=0;
    kdev_t dev=0;
    size_t size;
    struct kmouse_options options;
    struct kmouse_mouse *mPtr;
    struct kmouse_inproto *p;
    struct kmouse_protospec protospec;
    char devname[KMOUSE_DEVNAME_LEN];
    
    if (!suser())
	return -EPERM; /* only root (/etc/rc.d) can configure the device */

    /* Backward compatible: fold the old commands in the new ones */
    if (cmd == KMOUSE_OLD_GETVER) cmd = KMOUSE_IOC_GETVER;
    if (cmd == KMOUSE_OLD_SETOPT) cmd = KMOUSE_IOC_SETOPT;

    size = _IOC_SIZE(cmd);
    PDEBUGG("ioctl %x -- size %i\n",cmd, size);

    /* Ok, now verify the user area */
    if (_IOC_TYPE(cmd) != KMOUSE_IOC_MAGIC) return -EINVAL;
    if (_IOC_NR(cmd) < KMOUSE_IOC_MIN) return -EINVAL;
    if (_IOC_NR(cmd) > KMOUSE_IOC_MAX) return -EINVAL;
    if (_IOC_DIR(cmd) & _IOC_READ)
        err = verify_area(VERIFY_WRITE, (void *)arg, size);
    else if (_IOC_DIR(cmd) & _IOC_WRITE)
        err =  verify_area(VERIFY_READ, (void *)arg, size);
    if (err) return err;

    /* now the real work */
    switch(cmd) {

        case KMOUSE_IOC_GETVER:
            put_user((int)KMOUSE_VERSION_N,(int *)arg);
            break;

        case KMOUSE_IOC_SETOPT:
            memcpy_fromfs(&options,(void *)arg,size);
            dev = to_kdev_t(options.dev);
            PDEBUG("options for dev %s\n",kdevname(dev));
            mPtr = kmouse_lookfor(dev, 1);
            if (!mPtr) return -ENOMEM;

            if (options.protoid != PROTO_UNKNOWN) {
		PDEBUG("specified proto id is %i\n",options.protoid);
                for (p = kmouse_inprotolist; p; p = p->next)
                    if (p->protoid == options.protoid)
                        break;
                if (p) kmouse_mouse_init(mPtr, p->name);
            }
	    if (!mPtr->opt) {
		mPtr->opt = kmalloc(sizeof(options),GFP_KERNEL);
		if (!mPtr->opt)
		    return -ENOMEM;
	    }
	    *(mPtr->opt) = options;

            if ( (p=mPtr->proto) )
                PDEBUG("proto is %s (%i)\n",
                       p ? p->name : "unknown", p->protoid);
            break;

	case KMOUSE_IOC_SETDEV:
            memcpy_fromfs(devname,(void *)arg, size);
            PDEBUG("opening %s\n",devname);
            return kmouse_open_device(devname);

	case KMOUSE_IOC_SETPROTO:
            memcpy_fromfs(&protospec,(void *)arg, size);
            dev = to_kdev_t(protospec.device);
            mPtr = kmouse_lookfor(dev, 1);
            if (!mPtr) return -ENOMEM;
            PDEBUG("dev %s - protocol %s\n",kdevname(dev),protospec.proto);
            kmouse_mouse_init(mPtr, protospec.proto);
            return mPtr->proto ? 0 : -EINVAL;

        case KMOUSE_IOC_RESET:
            kmouse_close_devices();
            for (mPtr = kmouse_mlist; mPtr; mPtr = kmouse_mlist)
                kmouse_unlink(mPtr);
            break;

        case KMOUSE_IOC_SETXRANGE:
            PDEBUG("range x: %i\n",(int)arg);
            range_x = arg; break;

        case KMOUSE_IOC_SETYRANGE:
            PDEBUG("range y: %i\n",(int)arg);
            range_y = arg; break;

        default:
            printk("ioctl %x not implemented, yet\n", cmd);
            return -EINVAL;

        }
        
    return 0;
} 

/*
 * Select. The same function for all the nodes.
 */
int kmouse_select (struct inode *inode, struct file *filp,
                           int sel_type, select_table *wait)
{
    struct kmouse_client *client = filp->private_data;

    PDEBUGG("Select\n");
    if (sel_type == SEL_IN) {
        if (client->first || client->pending || kmouse_killed)
            return 1;
        select_wait(&client->queue,wait);
        PDEBUGG("select_wait\n");
        return 0;
    }
    if (sel_type == SEL_OUT)
        return 1; /* always writable */
    return 0; /* never exception-able */
}

/*
 * Write is easy: only the ctl node needs it. Other nodes just
 * eat everything
 */
int kmouse_write (struct inode *inode, struct file *filp,
                             const char *buf, int count)
{
    struct kmouse_client *client = filp->private_data;

    if (client->proto != KMOUSE_PROTO_GPM)
        return count;

     PDEBUGG("written by %p (%p)\n",client,filp);
     if (count < sizeof(Gpm_Connect))
         return -EINVAL;
     memcpy_fromfs(&client->conn,buf,sizeof(Gpm_Connect));
     return sizeof(Gpm_Connect);
}

/*
 * Read is the most difficult one: different data formats need
 * to be generated.
 */
int kmouse_read (struct inode *inode, struct file *filp,
                             char *buf, int count)
{
    struct kmouse_client *client = filp->private_data;
    Kmouse_Event *ptr;
    signed char *localbuf = client->buf; /* signed to calculate deltas */
    int bytes_read = 0;
    int zsize;

    if (kmouse_killed) return 0; /* close it up */

    if (!client->first && !client->pending && !client->zpending) {
        if (filp->f_flags & O_NONBLOCK)
            return -EAGAIN;
        interruptible_sleep_on(&client->queue);
        if (current->signal & ~current->blocked)
            return -ERESTARTSYS;
        if (kmouse_killed) return 0; /* make the client close it */
    }

    switch(client->proto) { /* different implementations */

      case KMOUSE_PROTO_GPM:
        #define QUANTUM sizeof(Gpm_Event)
        while (count >= QUANTUM) {
            Gpm_Event ev;
            int tmp;

            /* build the gpm event */
            ptr = client->first;
            kmouse_realxy(ptr);
            if (!filp->f_pos) {
                client->u.gpm.x = screenX;
                client->u.gpm.y = screenY;
            }
            tmp = ev.buttons = ptr->p.buttons;
            ev.modifiers = ptr->p.shift;
            ev.vc = ptr->vc;
            ev.x = screenX;
            ev.y = screenY;
            ev.dx = ev.x - client->u.gpm.x;
            ev.dy = ev.y - client->u.gpm.y;
            ev.type = ptr->type;
            ev.margin = ptr->margin;
            ev.clicks = ptr->clicks;
            if (ev.type & GPM_UP)
                ev.buttons ^= client->u.gpm.b; /* only the released one */
            client->u.gpm.b = tmp;
            client->u.gpm.x = ev.x;
            client->u.gpm.y = ev.y;

            memcpy_tofs(buf,&ev,QUANTUM);
            buf += QUANTUM; count -= QUANTUM; bytes_read += QUANTUM;
            client->first = ptr->next;
            if (ptr == client->last)
                client->last = NULL;
            client->nitems--;
            if (ptr->zptr)
                kfree(ptr->zptr);
            kfree(ptr);
        }
        #undef QUANTUM
        break;

      case KMOUSE_PROTO_MSC:
        while (client->pending && count) {
            put_user(localbuf[5-client->pending],buf);
            PDEBUGG("read leftover: one byte (pending was %i)\n",
                    client->pending);
            buf++; bytes_read++; client->pending--; count--;
        }

        /* then, process new packets */
        while(count && (ptr = client->first)) {
            int dx, dy, x, y;

            if (client->first->p.flags & KMOUSE_FLAG_ABS) {
                /*
                 * Map absolute to relative. Lock the pointer, and then
                 * push it: this means that huge movements count 0.
                 * First making delta and then subtracting makes the pointer
                 * move top-left. First scale then delta instead.
                 */
                x = (((ptr->p.x) >> 14) * range_x) >> 16;
                y = (((ptr->p.y) >> 14) * range_y) >> 16;
                dx = x - client->u.msc.x;
                dy = y - client->u.msc.y;
                client->u.msc.x = x;
                client->u.msc.y = y;
                PDEBUGG("abs: %i %i\n",dx,dy);
                /* zero if too big (more than 50)*/
                if ( (dx>50) || (dy>50) || (dx<-50) || (dy<-50) ) {
                    dx = 0; dy = 0;
                }
                /* and don't loop on the event */
                ptr->p.x = ptr->p.y = 0;
            } else { /* relative */
                dx = ptr->p.x; dy = ptr->p.y;
                PDEBUGG("loop: (%03i,%03i) B %i\n",
                        dx, dy, ptr->p.buttons);
                if (dx >  250) dx =  250;    /* clustering can be massive... */
                if (dx < -250) dx = -250;
                if (dy >  250) dy =  250;
                if (dy < -250) dy = -250;
                ptr->p.x -= dx;
                ptr->p.y -= dy;
            }
            localbuf[0] = 0x80 | ((ptr->p.buttons&0x07)^0x07);
            localbuf[3] =  dx - (localbuf[1] = dx/2); /* Markus */
            localbuf[4] = -dy - (localbuf[2] = -dy/2);
            if (count >= 5) { /* whole packet */
                memcpy_tofs(buf,localbuf,5);
            buf += 5; count  -= 5; bytes_read += 5;
            } else { /* part of a packet */
                client->pending=5;
                while (client->pending && count) {
                    put_user(localbuf[5 - client->pending],buf);
                    PDEBUGG("read partial: one byte\n");
                    buf++; bytes_read++; client->pending--; count--;
                }
            }
            if (ptr->p.x || ptr->p.y) {
                PDEBUG("looping: %i,%i left\n", ptr->p.x, ptr->p.y);
                continue; /* some dx/dy left */
            }
            client->first = ptr->next;
            if (client->last == ptr)
                client->last = NULL;
            client->nitems--;
            if (ptr->zptr)
                kfree(ptr->zptr);
            kfree(ptr);
        }
        break;

      case KMOUSE_PROTO_RNW: /* By now, only return sizeof(Kmouse_Packet) */
      case KMOUSE_PROTO_NEW:

        #define SIZE sizeof(Kmouse_Packet) /* the base size, Z is extra */
        count &= ~3; /* transfer multiples of 4 bytes */

        /*
         * First of all, pending bytes
         */
        if (client->pending && count) {
            PDEBUGG("reading %i pending bytes\n",client->pending);
            zsize = min(client->pending, count);
            memcpy_tofs(buf,localbuf+SIZE-client->pending,zsize);
            count -= zsize; buf += zsize; bytes_read += zsize;
            client->pending -= zsize;
        }
        /*
         * Then, zpending bytes
         */
        if (client->zpending && count) {
            PDEBUGG("reading %i zpending bytes\n",client->zpending);
            zsize = min(client->zpending, count);
            memcpy_tofs(buf,client->u.new.zbuf,zsize);
            count -= zsize; buf += zsize; bytes_read += zsize;
            client->zpending -= zsize; client->u.new.zbuf += zsize;
            if (!client->zpending)
                kfree(client->u.new.zptr);
        }
        /*
         * Then, process new packets. If read is partial save pending
         * and zpending bytes.
         */
        while(count && (ptr = client->first)) {
            zsize = sizeof(s32) *
                (kmouse_nr_dimensions((void *)ptr) - 2);
            if (count >= SIZE) { /* whole packet */
                memcpy_tofs(buf, &(ptr->p), SIZE);
		buf += SIZE; count  -= SIZE; bytes_read += SIZE;
		/*
                 * Ok, now try to push out Z as well
                 */
                if (zsize)
                    if (count >= zsize) { /* push all z's */
                        memcpy_tofs(buf, ptr->zptr, zsize);
                        buf += zsize; count -= zsize; bytes_read += zsize;
                    }
                    else { /* push partial z */
                        memcpy_tofs(buf, ptr->zptr, count);
                        client->zpending = zsize - count;
                        PDEBUGG("saving %i zpending bytes\n",client->zpending);
                        client->u.new.zptr = ptr->zptr;
                        ptr->zptr = NULL; /* don't free it */
                        client->u.new.zbuf = client->u.new.zptr + count;
                        buf += count; bytes_read += count; count = 0;
                    }
            } else { /* part of a packet: save pending and zpending */
                memcpy(localbuf,&(ptr->p),SIZE);
                memcpy_tofs(buf, localbuf, count);
                client->pending = SIZE - count;
                PDEBUGG("saving %i pending bytes\n",client->pending);
                buf += count; bytes_read += count; count = 0;
                /* save the z part as well */
                client->zpending = zsize;
                PDEBUGG("saving %i zpending bytes\n",client->zpending);
                client->u.new.zptr = ptr->zptr;
                ptr->zptr = NULL;
                client->u.new.zbuf = client->u.new.zptr;
            }

            client->first = ptr->next;
            if (client->last == ptr)
                client->last = NULL;
            client->nitems--;
            if (ptr->zptr)
                kfree(ptr->zptr);
            kfree(ptr);
            #undef SIZE
        }
        break;


      default:
        return -ENODEV;
    }

    filp->f_pos += bytes_read;
    return bytes_read;
}



/*===================================================== the devices */


struct file_operations kmouse_fops = {
    kmouse_seek,
    kmouse_read,
    kmouse_write,
    NULL,              /* readdir */
    kmouse_select,
    kmouse_ioctl,
    NULL,              /* mmap */
    kmouse_open,
    kmouse_release,
    NULL,              /* fsync */
    NULL,              /* fasync */
};


struct kmouse_misc_struct {struct miscdevice dev; int proto;};

struct kmouse_misc_struct kmouse_mice[] = {
    {{ MISC_DYNAMIC_MINOR , "kmouse",   &kmouse_fops}, KMOUSE_PROTO_MSC},
    {{ MISC_DYNAMIC_MINOR , "gpmctl",   &kmouse_fops}, KMOUSE_PROTO_GPM},
    {{ MISC_DYNAMIC_MINOR , "kpointer", &kmouse_fops}, KMOUSE_PROTO_NEW},
    {{ MISC_DYNAMIC_MINOR , "rpointer", &kmouse_fops}, KMOUSE_PROTO_RNW},
    {{ 0,                   NULL,       NULL        }, 0}
};

#define foreachnode(p) for (p=kmouse_mice; p->dev.name; p++)

static int minortoproto(int minor)
{
    struct kmouse_misc_struct *miscp;

    foreachnode(miscp)
        if (miscp->dev.minor == minor)
            break;
    return miscp->proto;
}

/*===================================================== /proc for debugging */
#if defined(KMOUSE_DEBUG) && LINUX_VERSION_CODE > VERSION_CODE(1,99,3)

int kmouse_read_proc(char *buf, char **start, off_t offset,
                   int len, int unused)
{
    struct kmouse_inproto *p;
    struct kmouse_mouse *mPtr;
    struct kmouse_client *client;
    struct kmouse_misc_struct *miscp;
    struct kmouse_device *dev;

    #define LIMIT (PAGE_SIZE-160) /* don't print anymore after this size */
    

    /*
     * Report the mice, the mPtrs, the clients, the /dev/nodes
     */
    len = sprintf(buf, "Registered protocols:\n");
    for (p = kmouse_inprotolist; p; p = p->next) {
        if (len>LIMIT) return len;
        len += sprintf(buf+len,"%8s (%i -- %02x %02x %02x %02x): "
                       "function @%p, usage %i\n",
                       p->name, p->howmany, p->proto[0], p->proto[1],
                       p->proto[2], p->proto[3], p->fun, p->usage);
    }

    len += sprintf(buf+len, "\nRegistered mice:\n");
    for (mPtr = kmouse_mlist; mPtr; mPtr = mPtr->next) {
        if (len>LIMIT) return len;
        len += sprintf(buf+len,"%s: proto %8s, opt %p, errors %i/%i\n",
                       kdevname(mPtr->dev), mPtr->proto->name, mPtr->opt,
                       mPtr->errors, mPtr->decoded);
    }

    len += sprintf(buf+len, "\nClients:\n");
    for (client = client_first; client; client = client->next) {
        if (len>LIMIT) return len;
        len += sprintf(buf+len,"%p: proto %04x, events %i, bytes %i\n",
                       client, client->proto, client->nitems, client->pending);
        len += sprintf(buf+len,"    modifiers: %x %x, evmask %x %x\n",
                       client->conn.minMod, client->conn.maxMod,
                       client->conn.eventMask, client->conn.defaultMask);
        /* FIXME -- print packet lists */
    }

    len += sprintf(buf+len, "\nData Structures:\n");
    if (len>LIMIT) return len;
    len += sprintf(buf+len, "client %i\n",sizeof(struct kmouse_client));
    if (len>LIMIT) return len;
    len += sprintf(buf+len, "mouse %i\n",sizeof(struct kmouse_mouse));
    if (len>LIMIT) return len;
    len += sprintf(buf+len, "event %i\n",sizeof(struct Kmouse_Event));
    if (len>LIMIT) return len;
    len += sprintf(buf+len, "inproto %i\n",sizeof(struct kmouse_inproto));
    if (len>LIMIT) return len;
    len += sprintf(buf+len, "device %i\n",sizeof(struct kmouse_device));
    
    len += sprintf(buf+len, "\nOpened Devices:\n");
    for (dev = kmouse_devicelist; dev; dev = dev->next) {
        if (len>LIMIT) return len;
        len += sprintf(buf+len, "%p: filp %p, inode %p\n",
                       dev,dev->filp,dev->inode);
    }

    len += sprintf(buf+len, "\nDevice Nodes:\n");
    foreachnode(miscp) {
        if (len>LIMIT) return len;
        len += sprintf(buf+len,"mknod -m 666 /dev/%s c %i %i\n",
		       miscp->dev.name, MISC_MAJOR, miscp->dev.minor);
    }
    return len;
}
#undef LIMIT

struct proc_dir_entry kmouse_proc_entry = {
        0,                 /* low_ino: the inode -- dynamic */
        6, "kmouse",       /* len of name and name */
        S_IFREG | S_IRUGO, /* mode */
        1, 0, 0,           /* nlinks, owner, group */
        0, NULL,           /* size - unused; operations -- use default */
        &kmouse_read_proc,  /* function used to read data */
        /* nothing more */
	};


#endif

/*===================================================== module stuff */

/*
 * Initialize yourself
 */

int init_module(void)
{
    int err;
    struct kmouse_misc_struct *miscp;
    extern void kmouse_init_protocols(void);
    
#ifndef KMOUSE_DEBUG
    static struct symbol_table kmouse_syms = {

    #include <linux/symtab_begin.h>
        X(kmouse_steal_char),
	X(kmouse_steal_packet),
        X(kmouse_register_inproto),
	X(kmouse_unregister_inproto),
    #include <linux/symtab_end.h>
    };

    register_symtab(&kmouse_syms);

#else if LINUX_VERSION_CODE > VERSION_CODE(1,99,3)
    proc_register_dynamic(&proc_root, &kmouse_proc_entry);
#endif
    
    kmouse_tqueue.routine = kmouse_dispatch;
    kmouse_tqueue.data = NULL; /* unused */
    
    /* register yourself -- four nodes */

    foreachnode(miscp) {
	if (minor)
            miscp->dev.minor = minor--;
        err = misc_register(&(miscp->dev));
        if (err) goto err_register;
    }

    /* the line discipline for serial mice */

    memset(&kmouse_ldisc, 0, sizeof(kmouse_ldisc));
    kmouse_ldisc.magic = TTY_LDISC_MAGIC;
    kmouse_ldisc.receive_room = kmouse_ldisc_room;
    kmouse_ldisc.receive_buf =  kmouse_ldisc_receive;
    kmouse_ldisc.open =         kmouse_ldisc_open;
    kmouse_ldisc.close =        kmouse_ldisc_close;
    err = tty_register_ldisc(N_MOUSE,&kmouse_ldisc);
    if (err) goto err_register;

    /* serial driver defaults from .../drivers/char/serial.c -- AU */
    kmouse_termios = tty_std_termios;
    kmouse_termios.c_cflag = B9600 | CS8 | CREAD | HUPCL | CLOCAL;

    kmouse_new = kmalloc(sizeof(struct kmouse_mouse), GFP_KERNEL);
    memset(kmouse_new,0,sizeof(struct kmouse_mouse));

    /*
     * Create a thread to handle nonatomic allocations, it accounts
     * for usage: count it now otherwise a "insmod; rmmod" would fault
     */
    MOD_INC_USE_COUNT;
    kernel_thread(kmouse_thread, NULL, /* FIXME -- all these clones? */
                  CLONE_PID|CLONE_FS|CLONE_VM|CLONE_FILES|CLONE_SIGHAND);

    kmouse_init_protocols();

    printk("kmouse " KMOUSE_VERSION " loaded\n");
    return 0;

    /*
     * Error: undo the incomplete registration
     */
    err_register: foreachnode(miscp) misc_deregister(&(miscp->dev));

    printk("kmouse: fatal error %i while registering\n",-err);

    return err;
}

void cleanup_module (void)
{
    struct kmouse_misc_struct *miscp;
    extern void kmouse_remove_protocols(void);

#if defined(KMOUSE_DEBUG) && LINUX_VERSION_CODE > VERSION_CODE(1,99,3)
    proc_unregister(&proc_root, kmouse_proc_entry.low_ino);
#endif

    if (kmouse_new) kfree(kmouse_new);
    foreachnode(miscp) misc_deregister(&(miscp->dev));
    tty_register_ldisc(N_MOUSE, NULL);
    /* no clients are there, because we count it as usage */

    kmouse_remove_protocols(); /* just in case :) */
    
    printk("kmouse " KMOUSE_VERSION " removed\n");
}

