/*
 * kmouse.h -- common information to kmouse and gpm configuring it
 *
 * Copyright (C) 1996 Alessandro Rubini (rubini@linux.it)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 ********/

#ifndef _KMOUSE_H_
#define _KMOUSE_H_

#ifndef _GPM_H_
#include <gpm.h>   /* still needs some gpm structures */
#endif

#define KMOUSE_VERSION   "0.47B"
#define KMOUSE_VERSION_N    47   /* decimal: MMMmmm */

#define KMOUSE_V_MAJOR(n) ((n)/1000)
#define KMOUSE_V_MINOR(n) ((n)%1000)

enum kmouse_proto {        /* this is the protoid field of kmouse_inproto */
    PROTO_UNKNOWN = 0,
    PROTO_BARE    = 1,
    PROTO_MSC     = 2,
    PROTO_SUN     = 3, 
    PROTO_PS2     = 4,
    PROTO_MS      = 5,     /* with three button extension */
    PROTO_LOGI    = 6,
    PROTO_MM      = 6,     /* the same as logi */
    PROTO_MMAN    = 7,
    PROTO_NCR     = 8,
    PROTO_WACOM   = 9
};

/*
 * The new extended protocol: it's 16 bytes minimum, but can be longer
 */
typedef struct Kmouse_Packet {
    __u16 magic;    /* two bytes in local-endian */
    __u16 flags;    /* the #buttons, #dim etc. */
    __u16 buttons;  /* up to 16 buttons available */
    __u16 shift;    /* the keyboard modifiers (kbd.h) */
    __s32 x;
    __s32 y;
    __s32 z[0];     /* extra coordinates */
} Kmouse_Packet;

#define KMOUSE_MAGIC         0xf00f
#define KMOUSE_SWAPPED_MAGIC 0x0ff0

/*
 * The flags: macros describe the bits, and inline
 * functions decode the field. Encoding is only done in the kernel
 */
#define KMOUSE_FLAG_NRDIM    0x000f   /* number of dimensions (2-15) */
#define KMOUSE_FLAG_NRBUTT   0x00f0   /* number of buttons (1-16) */
#define KMOUSE_FLAG_ABS      0x0100   /* absolute-pointing device */
#define KMOUSE_FLAG_Z_PRESS  0x0200   /* zed-is-pressure */
#define KMOUSE_FLAG_SPECIAL  0x8000   /* this is a special packet */

/*
 * Absolute pointing devices are scaled by the module: the minimum
 * value is 0, the maximum value is 1<<30-1. Out-of-screen values are
 * less-than-or-equal-to KMOUSE_OOS_LO, and greater-than-or-equal-to
 * KMOUSE_OOS_HI.
 */
#define KMOUSE_ABS_MAX (((__s32)1<<30)-1)
#define KMOUSE_ABS_MIN ((__s32)0)
#define KMOUSE_IS_OOS(value) ((__s32)(value)>>30)
#define KMOUSE_OOS_LO (KMOUSE_ABS_MIN-1)
#define KMOUSE_OOS_HI (KMOUSE_ABS_MAX+1)

extern inline int kmouse_packet_p(void *ptr)
{ return *(__u16 *)ptr == KMOUSE_MAGIC; }

extern inline int kmouse_swapped_packet_p(void *ptr)
{ return *(__u16 *)ptr == KMOUSE_SWAPPED_MAGIC; }

extern inline int kmouse_special_p(Kmouse_Packet *p)
{  return p->flags & KMOUSE_FLAG_SPECIAL; }

extern inline int kmouse_z_pressure_p(Kmouse_Packet *p)
{  return p->flags & KMOUSE_FLAG_Z_PRESS; }

extern inline int kmouse_absolute_p(Kmouse_Packet *p)
{  return p->flags & KMOUSE_FLAG_ABS; }

extern inline int kmouse_nr_dimensions (Kmouse_Packet *p)
{ return (p->flags & KMOUSE_FLAG_NRDIM); }

extern inline int kmouse_nr_buttons (Kmouse_Packet *p)
{ return 1 + ((p->flags & KMOUSE_FLAG_NRBUTT) >> 4); } /* 1 to 16 */

extern inline int kmouse_sizeof(Kmouse_Packet *p)
{ return sizeof(Kmouse_Packet) + sizeof(__s32) * (kmouse_nr_dimensions(p)-2); }




/*
 * This structure is passed from userspace to configure the device
 */
struct kmouse_options {
    dev_t dev;
    char sequence[8];
    unsigned short protoid, three;
    unsigned short delta, accel;
    unsigned short scale, scaley;
    unsigned short time, tap;
    unsigned short ptrdrag, bprev;
    unsigned short nochordmiddle;
};

/*
 * And this one is needed to tell a protocol for a device
 */
struct kmouse_protospec {
    dev_t device;
    char proto[16];
};

/*
 * The ioctl numbers: magic is 'k', sequence from 0xc0 up
 */
#define KMOUSE_IOC_MAGIC       'k'
#define KMOUSE_IOC_MIN                 0xc0
#define KMOUSE_IOC_GETVER     _IOR('k', 0xc0, int)
#define KMOUSE_IOC_SETOPT     _IOW('k', 0xc1, struct kmouse_options)
#define KMOUSE_IOC_SETDEV     _IOW('k', 0xc2, char[KMOUSE_DEVNAME_LEN])
#define KMOUSE_IOC_SETPROTO   _IOW('k', 0xc3, struct kmouse_protospec)
#define KMOUSE_IOC_RESET       _IO('k', 0xc4)
#define KMOUSE_IOC_SETXRANGE   _IO('k', 0xc5)
#define KMOUSE_IOC_SETYRANGE   _IO('k', 0xc6)
#define KMOUSE_IOC_MAX                 0xc6

#define KMOUSE_OLD_GETVER 0x6bc0
#define KMOUSE_OLD_SETOPT 0x6bc1

/* the filename for a device */
#define KMOUSE_DEVNAME_LEN 32

#define KMOUSE_B_LEFT   GPM_B_LEFT 
#define KMOUSE_B_MIDDLE GPM_B_MIDDLE
#define KMOUSE_B_RIGHT  GPM_B_RIGHT

#ifdef __KERNEL__

#undef PDEBUG
#undef PDEBUGG

#ifdef KMOUSE_DEBUG
#  define PDEBUG(fmt,args...) printk (KERN_DEBUG "kmouse: " fmt , ## args)
#  define Static
#else
#  define PDEBUG(fmt,args...)
#endif
#define PDEBUGG(fmt,args...)


/*
 * This is used by kmouse to grab data from the mouse drivers.
 * Allow for 4 dimensions -- should suffice for some time.
 */
extern void kmouse_steal_char(kdev_t dev, unsigned char datum);
extern void kmouse_steal_packet( kdev_t dev, int buttons, int flags,
			    int dx, int dy, int dz, int d4);

/*
 * The Kmouse_Event structure is used internally to manage data: it
 * encompasses the new protocol and fields specific to Gpm_Event.  It
 * should be no more than 32 bytes on the PC, to avoid memory leakage
 * (if clustering is disabled, a lot of events might reside
 * in the read queue of the device.  
 */
typedef struct Kmouse_Event {
    Kmouse_Packet p;
    s32 *zptr;
    struct kmouse_options *opt; /* relative to the calling mouse */
    struct Kmouse_Event *next;
    unsigned char vc, clicks;
    unsigned char type;   /* Gpm_Etype */
    unsigned char margin; /* Gpm_Margin */
} Kmouse_Event;

#define KMOUSE_BUFFER_LEN 32
struct kmouse_mouse {
    struct kmouse_mouse *next;
    struct kmouse_inproto *proto;
    struct kmouse_options *opt;
    void *private_data;
    int decoded;                    /* statistics... */
    short errors;                   /* more stats */
    short missing;                  /* before decoding */
    short head;
    short tail;
    kdev_t dev;                     /* sending device */
    char  data[KMOUSE_BUFFER_LEN];  /* head&tail point here */
    unsigned short xtrasize;        /* number of extra dimensions */
    u8 b1, b2;
    u16 w1, w2;                     /* these last fields are usable as      */
    u32 l1;                         /* private data to avoid a short malloc */
};



/*
 * The "inproto" is a mouse protocol. It has extra fields for
 * local status information, and three function pointers
 */

struct kmouse_inproto {
    char name[8];
    unsigned char proto[4];  /* identifying data */
    unsigned char extra[2];  /* how to tell an extra byte */
    u8 protoid;                  /* a few id's are renown, or use 0 */
    u8 howmany;                  /* bytes per packet */
    int (*fun)(Kmouse_Event *result, u8 *data, struct kmouse_mouse *self);
    int (*init)(struct kmouse_mouse *self);
    int (*cleanup)(struct kmouse_mouse *self);
    struct kmouse_inproto *next;
    int usage;
};


/*=================================================== inteface to inproto's */

extern int kmouse_register_inproto(struct kmouse_inproto *ptr);
extern int kmouse_unregister_inproto(struct kmouse_inproto *ptr);

#endif /* __KERNEL__ */

#endif /* _KMOUSE_H_ */
