/*
 * mouseconfig -- configure kmouse and let it run
 *
 * Copyright (C) 1996 Alessandro Rubini (rubini@linux.it)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 ********/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
 
#include <termios.h> /* TIOCGSERIAL etc. */

#include <linux/termios.h> /* line discipline */

#include "kmouse.h"

/*================================================== options for the module */

#define DEFAULT_THREE 0
#define DEFAULT_DELTA 25
#define DEFAULT_ACCEL 2
#define DEFAULT_SCALE 10
#define DEFAULT_TIME 250
#define DEFAULT_TAP 1
#define DEFAULT_PTRDRAG 1
#define DEFAULT_CHORDMIDDLE 0

#define DEFAULT_BAUD 1200

/*
 * The options structure is passed to the kernel. The protoid is
 * never used, but it was originally there...
 */

static struct kmouse_options default_opts = {
    0,                /* dev_t */
    "01234567",       /* sequence */
    0,                /* proto */
    DEFAULT_THREE,
    DEFAULT_DELTA,
    DEFAULT_ACCEL,
    DEFAULT_SCALE,
    DEFAULT_SCALE,
    DEFAULT_TIME,
    DEFAULT_TAP,
    DEFAULT_PTRDRAG,
    DEFAULT_CHORDMIDDLE
};

static struct kmouse_options current_opts, *opts;

/*================================================================ globals */

enum optionnames {DOUBLE, DRAG, SEQUENCE, DELTA, ACCEL, RESP, THREE, TAP,
                      CLEARLINE, BAUDRATE};

static int kmfd;
static int activefd=-1;
static dev_t activedev=0;
static int droplines;
static char devname[KMOUSE_DEVNAME_LEN];
static char *prgname;
static char *fname;
static int line;
static int opt_baud;

#define DEVICE "/dev/kmouse"


/*============================================================== baud rate */
/*
 * This function comes directly from gpm, which got it from selection.
 */
static int setspeed(int fd, int old, int new, unsigned short flags)
{
    struct termios tty;
    char *c;

    tcgetattr(fd, &tty);

    tty.c_iflag = IGNBRK | IGNPAR;
    tty.c_oflag = 0;
    tty.c_cflag = (tty.c_cflag & ~CSIZE) | CS8;
    tty.c_lflag = 0;
    tty.c_line = 0;
    tty.c_cc[VTIME] = 0;
    tty.c_cc[VMIN] = 1;

    switch (old) {
      case 9600:  tty.c_cflag = flags | B9600; break;
      case 4800:  tty.c_cflag = flags | B4800; break;
      case 2400:  tty.c_cflag = flags | B2400; break;
      case 1200:
      default:    tty.c_cflag = flags | B1200; break;
    }
    
    tcsetattr(fd, TCSAFLUSH, &tty);

    switch (new) {
      case 9600:  c = "*q";  tty.c_cflag = flags | B9600; break;
      case 4800:  c = "*p";  tty.c_cflag = flags | B4800; break;
      case 2400:  c = "*o";  tty.c_cflag = flags | B2400; break;
      case 1200:
      default:    c = "*n";  tty.c_cflag = flags | B1200; break;
    }

    write(fd, c, 2);
    usleep(100000);
    tcsetattr(fd, TCSAFLUSH, &tty);
    return 0;
}


/*==================================================== prepare serial ports */
/*
 * This function is called for every device being configured in
 */
void serial_init(int fd)
{
    int flags = CS8 | CSTOPB | CREAD | CLOCAL | HUPCL;
    int ldisc = N_TTY;
    int i;
    fd_set set; struct timeval timeout={0,0}; /* used when not debugging */

    if (ioctl(fd, TIOCSETD, &ldisc)) /* use N_TTY for setting features */
	return; /* failed: not a serial device */

    /* change from any available speed to the chosen one */
    for (i = 9600; i >= 1200; i /= 2)
	setspeed(fd, i, opt_baud, flags); /* FIXME -- flags and baudrate */

    #if 0 /* actually, I never understood this... */
    /* configure the sample rate */
    for (i=0;opt_sample<=sampletab[i].sample;i++)
	;
    write(fd,sampletab[i].code,1);
    #endif

    /*
     * flush any pending input (thanks, Miguel)
     */
    FD_ZERO(&set);
    while (1) {
	unsigned char c;
        FD_SET(fd,&set);
        switch(select(fd+1,&set,(fd_set *)NULL,(fd_set *)NULL,
		      &timeout/* zero */)) {
          case 1:  if (read(fd,&c,1)==0) break;
          case -1: continue;
	}
        break;
    }

    /*
     * reset the MouseMan/TrackMan to use the 3/4 byte protocol
     * (Stephen Lee, sl14@crux1.cit.cornell.edu)
     *
     * Hmmm... instead of "*X" write "*U" instead, this should bring
     * it to msc protocol.
     */
    #if 0 /* FIXME -- try to enable "*U" for normal mice */
    setspeed(fd, 1200, 1200, flags);
    write(fd, "*U", 2);
    setspeed(fd, 1200, opt_baud, flags);
    #endif

    ioctl(activefd, TIOCMGET, &flags);
    flags &= ~droplines;
    ioctl(activefd, TIOCMSET, &flags);

    ldisc = N_MOUSE;
    ioctl(fd, TIOCSETD, &ldisc); /* set line discipline to N_MOUSE */

}

static int tokenize(char *str, char **result, int maxfields)
{
    int i=0;

    while(i<maxfields) {
        result[i]=strtok(str," \t\n");
        if (!result[i]) return i;
        if (result[i][0]=='#')
            break;
        str=NULL;
        /* printf("tok %i: %s\n",i,result[i]); */
        i++;
    }
    return i;
}

struct command {
    char *name;
    int minf, maxf;
    int (*fn)(char **, int, struct command *);
    int argument;
};

int do_opt(char **tokens, int nf, struct command *self)
{
    struct winsize win;
    int i;
    static struct {char *in; char *out;} seq[] = {
        {"123","01234567"},
        {"132","02134657"},
        {"213","01452367"},
        {"231","02461357"},
        {"312","04152637"},
        {"321","04261537"},
        {NULL,NULL}
    };

    /* FIXME -- all the cfg variables */
    switch(self->argument) {

      case DOUBLE:
	opts->time = atoi(tokens[1]); break;

      case DRAG:
	opts->ptrdrag = atoi(tokens[1]); break;

      case SEQUENCE:
        for (i=0; seq[i].in && strcmp(seq[i].in,tokens[1]); i++)
            ;
        if (!seq[i].in) {
            fprintf(stderr,"%s:%i invalid sequence \"%s\"\n", fname, line,
                    tokens[1]);
            return 1;
        }
        strncpy(opts->sequence,seq[i].out,8);
        break;

      case DELTA:
	opts->delta = atoi(tokens[1]); break;

      case ACCEL:
	opts->accel = atoi(tokens[1]); break;

      case RESP:
        opts->scale = atoi(tokens[1]);
        if (!opts->scale) opts->scale = 100; /* maximum */
        else opts->scale = 100/opts->scale;  /* scale = inverse of resp.ness */
        i=open("/dev/tty0",O_RDONLY);
        ioctl(i, TIOCGWINSZ, &win);
        close(i);
        /*
         * the following operation is based on the observation that 80x50
         * has square cells. (An author-centric observation ;-)
         */
        opts->scaley = opts->scale * 50 * win.ws_col / 80 / win.ws_row;
        break;

      case THREE:
	opts->three = atoi(tokens[1]); break;
    
      case TAP:
	opts->tap = atoi(tokens[1]); break;

      case CLEARLINE:
        /*
         * Clear "dtr", "rts" or "both". Use activefd for this aim
         */
        if (activefd < 0) {
            fprintf(stderr,"%s:%i: Can't use \"clear\" in global section\n",
                    fname, line);
            return 1;
        }
        if (!strcmp(tokens[1],"dtr"))       droplines = TIOCM_DTR;
        else if (!strcmp(tokens[1],"rts"))  droplines = TIOCM_RTS;
        else if (!strcmp(tokens[2],"both")) droplines = TIOCM_DTR | TIOCM_RTS;
        else {
            fprintf(stderr,"%s:%i: Wrong token \"%s\": "
                    "use `dtr', `rts', or `both'\n", fname, line, tokens[1]);
            return 1;
        }
        break;

      case BAUDRATE:
	opt_baud = atoi(tokens[1]); break;

    }
    return 0;
}

int do_reset(char **tokens, int nf, struct command *self)
{
    if (ioctl(kmfd,KMOUSE_IOC_RESET,NULL)) {
        fprintf(stderr,"%s: reset %s: %s\n",prgname,DEVICE,strerror(errno));
        return 1;
    }
    return 0;
}


/*
 * This command prepares for another device. It closes the previous
 * device as well. Initialization is performed at close time
 */
int do_device(char **tokens, int nf, struct command *self)
{
    struct stat buf;
    int fd;

    if (activefd >= 0) {
        opts->dev = activedev;
	serial_init(activefd);
        ioctl(kmfd, KMOUSE_IOC_SETOPT, opts); /* push options out */
        close(activefd);
        activefd = -1;
        activedev = 0;
        devname[0] = '\0'; /* truncate it */
    }
    if (!nf) return 0;

    opt_baud = DEFAULT_BAUD; /* reset the default for the new device */
    strcpy(devname,tokens[1]);
    if ( (fd = open(devname, O_RDONLY))<0 ) {
        fprintf(stderr, "%s:%i: open(\"%s\"): %s\n", fname, line,
                devname, strerror(errno));
        return 1;
    }
    if (ioctl(kmfd, KMOUSE_IOC_SETDEV, devname)) {
        fprintf(stderr, "%s:%i: ioctl(setdev \"%s\"): %s\n", fname, line,
                devname, strerror(errno));
        close(fd);
        return 1;
    }

    activefd = fd;
    if (fstat(activefd, &buf)<0) {
        fprintf(stderr,"%s: stat(%s): %s\n",prgname,devname,strerror(errno));
        return 1;
    }
    activedev = buf.st_rdev;
    current_opts = default_opts;
    opts = &current_opts;

    return 0;    
}
    
int do_lut(char **tokens, int nf, struct command *self)
{
return 0; /* FIXME -- do lut loading */
}

int do_range(char **tokens, int nf, struct command *self)
{
    unsigned long x = atoi(tokens[1]);
   fprintf(stderr,"x is %li\n",x);
    ioctl(kmfd,self->argument,x);
    return 0;
}
    
int do_proto(char **tokens, int nf, struct command *self)
{
    struct kmouse_protospec spec;

    if (activefd < 0)
        return 1; /* Proto for whom? */
    spec.device = activedev;
    strcpy(spec.proto,tokens[1]);
    if (ioctl(kmfd, KMOUSE_IOC_SETPROTO, &spec)) {
        fprintf(stderr,"%s: setproto(%s): %s\n",prgname,spec.proto,
                strerror(errno));
        return 1;
    }
    return 0;        
}
  

struct command commands[] = {
    {"reset",       1,  1, do_reset, 0},
    {"LUT",         2, 10, do_lut, 0},
    {"device",      2,  2, do_device, 0},
    {"protocol",    2,  2, do_proto, 0},
    {"doubleclick", 2,  2, do_opt, DOUBLE},
    {"ptrdrag",     2,  2, do_opt, DRAG},
    {"buttonsequence",2,2, do_opt, SEQUENCE},
    {"delta",       2,  2, do_opt, DELTA},
    {"accel",       2,  2, do_opt, ACCEL},
    {"responsiveness",2,2, do_opt, RESP},
    {"threebuttons",2,  2, do_opt, THREE},
    {"tapbutton",   2,  2, do_opt, TAP},
    {"clear",       2,  2, do_opt, CLEARLINE},
    {"baudrate",    2,  2, do_opt, BAUDRATE},
    {"xrange",      2,  2, do_range, KMOUSE_IOC_SETXRANGE},
    {"yrange",      2,  2, do_range, KMOUSE_IOC_SETYRANGE},
    {NULL,0,0,NULL,0}
};

int main(int argc, char **argv)
{
    int nf, retval=0;
    FILE *f;
    char s[128];
    char *tokens[10];
    struct command *cptr;
    char *fname = "/etc/mouse.conf";

    prgname = argv[0];

    switch(argc) {
      case 2:
        fname = argv[1];
      case 1:
        break;
      default:
        fname = "/dev/null";
        fprintf(stderr,"%s: Wrong number of arguments\n",argv[0]);
        retval++;
    }

    if (!strcmp(fname,"-kill")) { /* FIXME -- horrible */
	system ("killall kmouse");
	sleep(1);
	system ("rmmod kmouse");
	exit(0);
    }

    if (!strcmp(fname,"-v") ||
	!strcmp(fname,"--version")) {
	fprintf(stderr,"%s version %s\n",argv[0],KMOUSE_VERSION);
	kmfd = open(DEVICE, O_RDWR);
	if (kmfd<0) {
	    fprintf(stderr,"%s: %s\n",DEVICE,strerror(errno));
	    exit(1);
	}
	if (ioctl(kmfd, KMOUSE_IOC_GETVER, &nf)) {
	    fprintf(stderr, "%s: ioctl(): %s\n",
		    DEVICE,strerror(errno));
	    exit(1);
	}
	fprintf(stderr,"Kmouse version %i.%02i\n",
	       KMOUSE_V_MAJOR(nf), KMOUSE_V_MINOR(nf));
	exit(0);
    }

    if (!(f=fopen(fname,"r"))) {
        fprintf(stderr,"%s: %s: %s\n",argv[0],fname,strerror(errno));
        retval++;
    }

    kmfd=open(DEVICE, O_RDWR);
    if (kmfd<0) {
        fprintf(stderr,"%s: %s: %s\n",argv[0],DEVICE,strerror(errno));
        retval++;
    } else if (ioctl(kmfd, KMOUSE_IOC_GETVER, &nf)) {
        fprintf(stderr, "%s: %s: ioctl(): %s\n",
                prgname,DEVICE,strerror(errno));
        retval++;
    } else if (nf!=KMOUSE_VERSION_N) {
        fprintf(stderr, "%s: %s: Version mismatch: "
                "module is %i.%i, program is %s\n", prgname, DEVICE,
                KMOUSE_V_MAJOR(nf), KMOUSE_V_MINOR(nf), KMOUSE_VERSION);
        retval++;
    }
    if (retval) return retval;

    opts = &default_opts;
    
    line = 0;
    while (fgets(s,128,f)) {
        line++;
        if ( (nf=tokenize(s,tokens,10)) <1) continue;
        
        for (cptr = commands; cptr->name; cptr++)
            if (!strcmp(tokens[0],cptr->name)) {
                if (nf < cptr->minf)
                    fprintf(stderr,"%s:%i: too few arguments to \"%s\"\n",
                            fname,line,tokens[0]);
                else if (nf > cptr->maxf)
                    fprintf(stderr,"%s:%i: too many arguments to \"%s\"\n",
                            fname,line,tokens[0]);
                else if (cptr->fn)
                    retval += cptr->fn(tokens,nf,cptr);
                break;
            }

        if (!cptr->name)
            fprintf(stderr,"%s:%i uknown keyword %s\n",
                    fname, line, tokens[0]);
    }

    /*
     * Ok, do configure the last device
     */
    retval += do_device(NULL,0,NULL);

    exit(retval);
}
