/*
 * protocols for kmouse'd devices
 *
 * Copyright (C) 1996 Alessandro Rubini (rubini@ipvvis.unipv.it)
 *
 * Several fixes by Markus Gutschke (gutschk@math.uni-muenster.de)
 * Good suggestions by Aaron Ucko (ucko@vax1.rockhurst.edu)
 *
 * Part of protocol decoding comes from selection, by Andrew Haylett
 * The idea of the kernel module is by Peter Orbaek
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 ********/


#ifndef __KERNEL__
#define __KERNEL__
#endif

#include <linux/types.h>   /* ulong and friends */
#include <linux/malloc.h>  /* kmalloc, kfree */
#include <linux/kdev_t.h>  /* building majors */

#include "kmouse.h"
#include "kmouse-config.h"

#ifdef CONFIG_PS2_MODULE
#  define CONFIG_PS2_PROTOCOL
#endif

/*
 * Each "kmouse_inproto" registers its own data structure at init time.
 *
 * This file is linked to core.c to build kmouse.o, but I'll provide
 * sample dynamic protocols.
 *
 * How does this work?
 *
 * Each protocol  is  described   by the "kmouse_inproto"   structure,
 * defined in "kmouse.h". The "proto" field embeds four bytes that can
 * help in identifying the first byte in a packet.  Look at core.c (or
 * gpm.c) to  look how they are  used. "extra" are  two bytes mask and
 * *value) used to identify an extra byte that modifies the meaning of
 * a packet: using "{0xff, 0x00}"  will ignore extra bytes.  "howmany"
 * tells how many  bytes build up  a packet. "protoid" is not actually
 * used: 0 will work.
 *
 * The "init"   function returns the number  of   extra dimensions (in
 * additions  to X  and Y)  that are  reported by the  pointer.  It is
 * passed  a "kmouse_mouse *" pointer in  order  to initialize its own
 * data structure.  The   following fields  exist  to be   used by the
 * protocol decoder:
 *
 *        void *private_data;
 *        u8 b1, b2;
 *        u16 w1, w2;
 *        u32 l1;
 *
 * If   "private_data" is  allocated   with  "kmalloc()",  the cleanup
 * function  should free  it. The  other  fields only  exist  to avoid
 * unnecessary allocations  for small  amount  of status  data.  Using
 * static variables to host status informatio is not correct because a
 * workstation  can    have two or  more     devices speaking the same
 * protocol, but each of them needs its own status information.
 *
 * The "cleanup"  function is    called  before a pointer    device is
 * released.
 *
 * "fun" is the actual decoder: it  should fill the p.buttons, p.flags
 * p.x and p.y fields.  If more than  two dimensions are reported, the
 * decoder should *atomically* allocate space for  the z dimensions in
 * "eptr->zptr".  The decoder  function should  return  the  number of
 * bytes it used up from  the "data" array. If  allocation of z fails,
 * the packet should be discarded. Discarding a packet is accomplished
 * by  returning the negative  of the number  of bytes that build up a
 * packet.
 *
 * A   protocol     is     registered  in       kmouse   by    calling
 * "kmouse_register_inproto()"    and      is      removed     through
 * "kmouse_unregister_inproto()". If you  plan  to add a new  protocol
 * decoder and you  have some problems feel free  to bug me. I hope to
 * offer an example of loadable protocol decoder in the future.
 * */

/*===================================================== the bare ms proto */
/*
 * This needs no initialization, because there's no private_data used,
 * nor extra size to for the packet to be reported
 */

#ifdef CONFIG_MAIN_PROTOCOLS    
int M_bare(Kmouse_Event *eptr,  u8 *data, struct kmouse_mouse *mPtr)
{
    /*
     * mPtr->b1 is "prevbuttons"
     * mPtr->b2 is 0 (default for two buttons, or 16 for three)
     */
    
    /* FIXME -- whatabout chordmiddle? */

    eptr->p.buttons =  ((data[0] & 0x20) >> 3) | ((data[0] & 0x10) >> 4);
    eptr->p.x=         (char)(((data[0] & 0x03) << 6) | (data[1] & 0x3F));
    eptr->p.y=         (char)(((data[0] & 0x0C) << 4) | (data[2] & 0x3F));
    /*
     * As long as there is no native "mman" support for "kmouse", we can
     * resort to using the bare microsoft protocol with chord-middle
     * behavior enabled; this requires passing "-o dtr" to "gpm";
     * M. Gutschke
     */
    #define GPM_B_BOTH (GPM_B_LEFT|GPM_B_RIGHT)
    if ( ((mPtr->b1^eptr->p.buttons) & GPM_B_BOTH)==GPM_B_BOTH ) {
	mPtr->b2 = 0x10; /* three are there */
        eptr->p.buttons = eptr->p.buttons ? GPM_B_MIDDLE : 0;
    }
    mPtr->b1 = eptr->p.buttons;
    
    eptr->p.flags = 0x0012 + mPtr->b2; /* 2/3 buttns, 2 dimensions, relative */
    return 3;
}

struct kmouse_inproto M_bare_proto = {
    "ms", {0x40, 0x40, 0x40, 0x00}, {0x00, 0xff},
    PROTO_BARE, 3,
    M_bare, NULL, NULL,
};
#endif /* CONFIG_MAIN_PROTOCOLS */

/*===================================================== the msc 3b proto */
/*
 * This needs no initialization, because there's no private_data used
 */

#ifdef CONFIG_MAIN_PROTOCOLS    
int M_msc(Kmouse_Event *eptr, u8 *data, struct kmouse_mouse *mPtr)
{
    eptr->p.flags = 0x022; /* 1+2 buttons, 2 dimensions */
    eptr->p.buttons= (~data[0]) & 0x07;
    eptr->p.x=      (char)(data[1]) + (char)(data[3]);
    eptr->p.y=     -((char)(data[2]) + (char)(data[4]));
    return 5;
}

struct kmouse_inproto M_msc_proto = {
    "msc", {0xf8, 0x80, 0x00, 0x00}, {0x00, 0xff}, 
    PROTO_MSC, 5,
    M_msc, NULL, NULL,
};
#endif /* CONFIG_MAIN_PROTOCOLS */


/*===================================================== sun (?) */

#ifdef CONFIG_OTHER_PROTOCOLS    
int M_sun(Kmouse_Event *eptr, u8 *data, struct kmouse_mouse *mPtr)
{
    eptr->p.flags = 0x022;
    eptr->p.buttons= (~data[0]) & 0x07;
    eptr->p.x=      (char)(data[1]);
    eptr->p.y=     -(char)(data[2]);
    return 3;
}

struct kmouse_inproto M_sun_proto = {
    "sun", {0xf8,0x00,0x0,0x0}, {0x00,0xff},
    PROTO_SUN, 3,
    M_sun, NULL, NULL,
};
#endif /* CONFIG_OTHER_PROTOCOLS */


/*===================================================== logitech */

#ifdef CONFIG_OTHER_PROTOCOLS    
int M_logi(Kmouse_Event *eptr, u8 *data, struct kmouse_mouse *mPtr)
{
    eptr->p.flags = 0x22;
    eptr->p.buttons= data[0] & 0x07;
    eptr->p.x=      (data[0] & 0x10) ?   data[1] : - data[1];
    eptr->p.y=      (data[0] & 0x08) ? - data[2] :   data[2];
    return 3;
}
    
struct kmouse_inproto M_logi_proto = {
    "logi", {0xe0,0x80,0x80,0x00}, {0x00,0xff},
    PROTO_LOGI, 3,
    M_logi, NULL, NULL,
};
#endif /* CONFIG_OTHER_PROTOCOLS */

/*===================================================== ps2 mouse */

#ifdef CONFIG_PS2_PROTOCOL
int M_ps2(Kmouse_Event *eptr, u8 *data, struct kmouse_mouse *mPtr)
{
    eptr->p.flags = 0x22;
    eptr->p.buttons=
        !!(data[0]&1) * GPM_B_LEFT +
        !!(data[0]&2) * GPM_B_RIGHT +
        !!(data[0]&4) * GPM_B_MIDDLE;

    eptr->p.x=   (data[0] & 0x10) ? data[1]-256 : data[1];
    eptr->p.y= -((data[0] & 0x20) ? data[2]-256 : data[2]);
    return 3;
}

struct kmouse_inproto M_ps2_proto = {
    "ps2", {0xc0,0x00,0x00,0x00}, {0x00,0xff},
    PROTO_PS2, 3,
    M_ps2, NULL, NULL,
};
#endif /* CONFIG_PS2_PROTOCOL */

/*===================================================== wacom tablet */
/*
 * Well, I'm interested in this stuff: it has both three dimensions
 * and absolute coordinates.
 */

#ifdef CONFIG_WACOM_PROTOCOL
int M_wacom_init(struct kmouse_mouse *mPtr)
{
    mPtr->w1 = 0; /* the maximum y */
    mPtr->l1 = 0; /* the maximum x */
    mPtr->b1 = 8; /* pressure threshold */
    mPtr->opt->three = 1; /* we do have the middle button */
    return 1; /* one extra dimension */
}

int M_wacom(Kmouse_Event *eptr, u8 *data, struct kmouse_mouse *mPtr)
{
    int x,y,z;
    
    /*
     * I just don't understand "A0": it is genetated on release,
     * but it is a whole packet, not a single byte. Correct?
     */
#if 0
    if (data[0] == 0xa0) { /* `release' is one byte only. Well: discard it. */
	PDEBUG("data: %02x %02x %02x %02x %02x %02x %02x\n",
               data[0], data[1], data[2], data[3], data[4], data[5], data[6]);
        return -1;
    }
#endif

    x = (data[1]<<7) | (data[2]&0x7f);
    y = (data[4]<<7) | (data[5]&0x7f);
    z = data[6]-64 + 128*(data[6]<64);

    if (x > mPtr->l1) { mPtr->l1 = x; PDEBUGG("maxx is %i\n",x);}
    if (y > mPtr->w1) { mPtr->w1 = y; PDEBUGG("maxy is %i\n",y);}
    PDEBUGG("xyz: %i, %i, %i\n",x,y,z);

    eptr->p.flags = 0x23 | KMOUSE_FLAG_ABS | KMOUSE_FLAG_Z_PRESS;

    eptr->p.buttons = (GPM_B_MIDDLE * !!(data[3]&0x10)); /* middle */
    if (!eptr->p.buttons)
        if (data[3]&0x20) eptr->p.buttons = (z > mPtr->b1) * GPM_B_RIGHT;
        else              eptr->p.buttons = (z > mPtr->b1) * GPM_B_LEFT;

    /*
     * Since the precision is well below 16 bits, do 16 bits calculation
     * to avoid overflow, and then shift.
     */
    eptr->p.x = ((x<<16) / mPtr->l1) << 14;
    eptr->p.y = ((y<<16) / mPtr->w1) << 14;
    eptr->zptr = kmalloc(sizeof(__s32), GFP_ATOMIC);
    if (!eptr->zptr) return -7;
    eptr->zptr[0] = ((z<<16) / 128) << 14; /* 128 is the reknown z-max */

    if (z > mPtr->b1) mPtr->b1 = 1; /* shmitt-trigger the threshold */
    else mPtr->b1 = 8;

    return 7;
}

struct kmouse_inproto M_wacom_proto = {
    "wacom", {0x80, 0x80, 0x80, 0x00}, {0x00, 0xff},
    PROTO_WACOM, 7,
    M_wacom, M_wacom_init, NULL,
    };
#endif /* CONFIG_WACOM_PROTOCOL */

/*========================================================= init/clean */

void kmouse_init_protocols(void)
{
#ifdef CONFIG_MAIN_PROTOCOLS
    kmouse_register_inproto(&M_bare_proto);
    kmouse_register_inproto(&M_msc_proto);
#endif
#ifdef CONFIG_OTHER_PROTOCOLS
    kmouse_register_inproto(&M_sun_proto);
    kmouse_register_inproto(&M_logi_proto);
#endif
#ifdef CONFIG_PS2_PROTOCOL
    kmouse_register_inproto(&M_ps2_proto);
#endif
#ifdef CONFIG_WACOM_PROTOCOL
    kmouse_register_inproto(&M_wacom_proto);
#endif
}

void kmouse_remove_protocols(void)
{
#ifdef CONFIG_MAIN_PROTOCOLS
    kmouse_unregister_inproto(&M_bare_proto);
    kmouse_unregister_inproto(&M_msc_proto);
#endif
#ifdef CONFIG_OTHER_PROTOCOLS
    kmouse_unregister_inproto(&M_sun_proto);
    kmouse_unregister_inproto(&M_logi_proto);
#endif
#ifdef CONFIG_PS2_PROTOCOL
    kmouse_unregister_inproto(&M_ps2_proto);
#endif
#ifdef CONFIG_WACOM_PROTOCOL
    kmouse_unregister_inproto(&M_wacom_proto);
#endif
}

















/*===================================================== other protocols */



#if 0 /* FIXME -- the other protocols */

/*
 * These are copied from the old kmouse, and need to be ported to the new
 * interface
 */


/* originally by M. Gutschke, modified by me, minor bug fix by MG :-) */
int M_mman(Gpm_Event *eptr,  unsigned char *data)
{
    static short bprev, nochordmiddle;


    if ( (data[0] & 0x40) == 0x00) { /* got the extra byte */
        nochordmiddle = 1;
	/* discard empty packages as they *seriously* confuse the
	   clustering code! */
	if (!(data[0] & 0x20) == !(bprev & GPM_B_MIDDLE))
	  return(1);
        if (data[0] & 0x20)
            bprev |= GPM_B_MIDDLE;
        else
            bprev &= ~GPM_B_MIDDLE;
        state->buttons = bprev;
        state->dx = state->dy = 0; /* no motion */
        return 0;
    }

    /* this is the start of a new packet */
    state->buttons= ((data[0] & 0x20) >> 3) | ((data[0] & 0x10) >> 4) |
                     (bprev & GPM_B_MIDDLE);
    state->dx=      (char)(((data[0] & 0x03) << 6) | (data[1] & 0x3F));
    state->dy=      (char)(((data[0] & 0x0C) << 4) | (data[2] & 0x3F));

    if (!nochordmiddle) {
        /*
         * chord middle reporting is probably not neccessary
         * if mouse knows about 3/4 protocol
         */
#define GPM_B_BOTH (GPM_B_LEFT|GPM_B_RIGHT)
        if ( ((bprev^state->buttons) & GPM_B_BOTH) == GPM_B_BOTH )
            state->buttons = state->buttons ? GPM_B_MIDDLE : 0; 
    }
    bprev=state->buttons;
    return 0;
}


/*
 * Busmice return a whole packet at a time, and are all equal.
 * For serial mice, autodetection works for: bare/mman, msc, ps2, logi/mm
 * ms with third button and sun can't be detected.
 */                          

struct kmouse_protocol protocols[]={
   { PROTO_MMAN,     3, {0x40, 0x40, 0x40, 0x00}, M_mman, {0x40, 0x00}, 0},
   { PROTO_BARE,     3, {0x40, 0x40, 0x40, 0x00}, M_bare, {0x00, 0xff}, 0},
   { PROTO_MSC,      5, {0xf8, 0x80, 0x00, 0x00}, M_msc,  {0x00, 0xff}, 0},
   { PROTO_SUN,      3, {0xf8, 0x80, 0x00, 0x00}, M_sun,  {0x00, 0xff}, 0},
   { PROTO_LOGI,     3, {0xe0, 0x80, 0x80, 0x00}, M_logi, {0x00, 0xff}, 0},
   { PROTO_PS2,      3, {0xc0, 0x00, 0x00, 0x00}, M_ps2,  {0x00, 0xff}, 1},
#if 0
   { PROTO_MS,       3, {0x40, 0x40, 0x40, 0x00}, M_ms,   {0x00, 0xff}, 0},
   { PROTO_WACOM     7, {0x80, 0x80, 0x80, 0x00}, M_wacom,{0x00, 0xff}, 0},
#endif
   { 0,              0, {0x00, 0x00, 0x00, 0x00}, NULL,   {0x00, 0xff}, 0}
};


#endif
